# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# Modified Amritpal Singh <amrit3701@gmail.com> on 07-07-2017

__title__ = "FreeCAD Rebar"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchRebar
#  \ingroup ARCH
#  \brief The Rebar object and tools
#
#  This module provides tools to build Rebar objects.
#  Rebars (or Reinforcing Bars) are metallic bars placed
#  inside concrete structures to reinforce them.

import FreeCAD
import ArchCommands
import ArchIFC
import ArchComponent
import Draft
import DraftVecUtils

from draftutils import params

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate

    # TODO: check if this import is still needed, and if so, whether
    # it can be moved made conditional on the GUI being loaded
    # for Rebar addon compatibility
    from bimcommands import BimRebar

    _CommandRebar = BimRebar.Arch_Rebar
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _Rebar(ArchComponent.Component):
    "A parametric reinforcement bar (rebar) object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Rebar"
        self.setProperties(obj)
        obj.IfcType = "Reinforcing Bar"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Diameter" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Diameter",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "The diameter of the bar"),
                locked=True,
            )
        if not "OffsetStart" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "OffsetStart",
                "Rebar",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The distance between the border of the beam and the first bar (concrete cover).",
                ),
                locked=True,
            )
        if not "OffsetEnd" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "OffsetEnd",
                "Rebar",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The distance between the border of the beam and the last bar (concrete cover).",
                ),
                locked=True,
            )
        if not "Amount" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "Amount",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "The amount of bars"),
                locked=True,
            )
        if not "Spacing" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Spacing",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "The spacing between the bars"),
                locked=True,
            )
            obj.setEditorMode("Spacing", 1)
        if not "Distance" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Distance",
                "Rebar",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The total distance to span the rebars over. Keep 0 to automatically use the host shape size.",
                ),
                locked=True,
            )
        if not "Direction" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "Direction",
                "Rebar",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The direction to use to spread the bars. Keep (0,0,0) for automatic direction.",
                ),
                locked=True,
            )
        if not "Rounding" in pl:
            obj.addProperty(
                "App::PropertyFloat",
                "Rounding",
                "Rebar",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The fillet to apply to the angle of the base profile. This value is multiplied by the bar diameter.",
                ),
                locked=True,
            )
        if not "PlacementList" in pl:
            obj.addProperty(
                "App::PropertyPlacementList",
                "PlacementList",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "List of placement of all the bars"),
                locked=True,
            )
        if not "Host" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Host",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "The structure object that hosts this rebar"),
                locked=True,
            )
        if not "CustomSpacing" in pl:
            obj.addProperty(
                "App::PropertyString",
                "CustomSpacing",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "The custom spacing of rebar"),
                locked=True,
            )
        if not "Length" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "Length",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "Length of a single rebar"),
                locked=True,
            )
            obj.setEditorMode("Length", 1)
        if not "TotalLength" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "TotalLength",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "Total length of all rebars"),
                locked=True,
            )
            obj.setEditorMode("TotalLength", 1)
        if not "Mark" in pl:
            obj.addProperty(
                "App::PropertyString",
                "Mark",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "The rebar mark"),
                locked=True,
            )

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def loads(self, state):

        self.Type = "Rebar"

    def getBaseAndAxis(self, wire):
        "returns a base point and orientation axis from the base wire"
        import DraftGeomUtils

        if wire:
            e = wire.Edges[0]
            # v = DraftGeomUtils.vec(e).normalize()
            v = e.tangentAt(e.FirstParameter)
            return e.Vertexes[0].Point, v
        if obj.Base:
            if obj.Base.Shape:
                if obj.Base.Shape.Wires:
                    e = obj.Base.Shape.Wires[0].Edges[0]
                    # v = DraftGeomUtils.vec(e).normalize()
                    v = e.tangentAt(e.FirstParameter)
                    return e.Vertexes[0].Point, v
        return None, None

    def getRebarData(self, obj):

        # if not obj.Host:
        #    return
        # if Draft.getType(obj.Host) != "Structure":
        #    return
        # if not obj.Host.Shape:
        #    return
        if not obj.Base:
            return
        if not obj.Base.Shape:
            return
        if not obj.Base.Shape.Wires:
            return
        if not obj.Diameter.Value:
            return
        if not obj.Amount:
            return
        father = None
        if obj.Host:
            father = obj.Host
        wire = obj.Base.Shape.Wires[0]
        axis = None
        if Draft.getType(obj.Base) == "Wire":  # Draft Wires can have "wrong" placement
            import DraftGeomUtils

            axis = DraftGeomUtils.getNormal(obj.Base.Shape)
        if not axis:
            axis = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(0, 0, -1))
        if hasattr(obj, "Direction"):
            if not DraftVecUtils.isNull(obj.Direction):
                axis = FreeCAD.Vector(obj.Direction)
                axis.normalize()
        size = 0
        if father:
            size = (ArchCommands.projectToVector(father.Shape.copy(), axis)).Length
        if hasattr(obj, "Distance"):
            if obj.Distance.Value:
                size = obj.Distance.Value
        if hasattr(obj, "Rounding"):
            if obj.Rounding:
                radius = obj.Rounding * obj.Diameter.Value
                import DraftGeomUtils

                wire = DraftGeomUtils.filletWire(wire, radius)
        wires = []
        if obj.Amount == 1:
            if size and father:
                offset = DraftVecUtils.scaleTo(axis, size / 2)
            else:
                offset = FreeCAD.Vector()
            wire.translate(offset)
            wires.append(wire)
        else:
            if obj.OffsetStart.Value:
                baseoffset = DraftVecUtils.scaleTo(axis, obj.OffsetStart.Value)
            else:
                baseoffset = None
            if hasattr(obj, "RebarShape") and obj.RebarShape == "Stirrup":
                interval = size - (obj.OffsetStart.Value + obj.OffsetEnd.Value + obj.Diameter.Value)
            else:
                interval = size - (obj.OffsetStart.Value + obj.OffsetEnd.Value)
            interval = interval / (obj.Amount - 1)
            vinterval = DraftVecUtils.scaleTo(axis, interval)
            for i in range(obj.Amount):
                if i == 0:
                    if baseoffset:
                        wire.translate(baseoffset)
                    wires.append(wire)
                else:
                    wire = wire.copy()
                    wire.translate(vinterval)
                    wires.append(wire)
        return [wires, obj.Diameter.Value / 2]

    def onChanged(self, obj, prop):
        if prop == "IfcType":
            root = ArchIFC.IfcProduct()
            root.setupIfcAttributes(obj)
            root.setupIfcComplexAttributes(obj)
        elif prop == "Host":
            if hasattr(obj, "Host"):
                if obj.Host:
                    # mark host to recompute so it can detect this object
                    obj.Host.touch()

    def execute(self, obj):

        if self.clone(obj):
            return
        if not self.ensureBase(obj):
            return
        if not obj.Base:
            # let pass without error so that object can receive a shape directly
            # FreeCAD.Console.PrintError(
            #    "No Base, return without a rebar shape for {}.\n"
            #    .format(obj.Name)
            # )
            return
        if not hasattr(obj.Base, "Shape") or (not obj.Base.Shape) or obj.Base.Shape.isNull():
            FreeCAD.Console.PrintError(
                "No Shape in Base, return without a rebar shape for {}.\n".format(obj.Name)
            )
            return
        if obj.Base.Shape.Faces:
            FreeCAD.Console.PrintError(
                "Faces in Shape of Base, return without a rebar shape for {}.\n".format(obj.Name)
            )
            return
        if not obj.Base.Shape.Edges:
            FreeCAD.Console.PrintError(
                "No Edges in Shape of Base, return without a rebar shape for {}.\n".format(obj.Name)
            )
            return
        if not obj.Diameter.Value:
            FreeCAD.Console.PrintError(
                "No Diameter Value, return without a rebar shape for {}.\n".format(obj.Name)
            )
            return
        if not obj.Amount:
            FreeCAD.Console.PrintError(
                "No Amount, return without a rebar shape for {}.\n".format(obj.Name)
            )
            return
        father = obj.Host
        fathershape = None
        if not father:
            # support for old-style rebars
            if obj.InList:
                if hasattr(obj.InList[0], "Armatures"):
                    if obj in obj.InList[0].Armatures:
                        father = obj.InList[0]
        if father:
            if hasattr(father, "Shape"):
                fathershape = father.Shape

        import Part

        # corner cases:
        #    compound from more Wires
        #    compound without Wires but with multiple Edges
        # Does they make sense? If yes handle them.
        # Does it makes sense to handle Shapes with Faces or even Solids?
        if not obj.Base.Shape.Wires and len(obj.Base.Shape.Edges) == 1:
            wire = Part.Wire(obj.Base.Shape.Edges[0])
        else:
            wire = obj.Base.Shape.Wires[0]
        if hasattr(obj, "Rounding"):
            # print(obj.Rounding)
            if obj.Rounding:
                radius = obj.Rounding * obj.Diameter.Value
                from DraftGeomUtils import filletWire

                wire = filletWire(wire, radius)
        bpoint, bvec = self.getBaseAndAxis(wire)
        if not bpoint:
            return
        axis = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(0, 0, -1))
        if fathershape:
            size = (ArchCommands.projectToVector(fathershape.copy(), axis)).Length
        else:
            size = 1
        if hasattr(obj, "Direction"):
            if not DraftVecUtils.isNull(obj.Direction):
                axis = FreeCAD.Vector(obj.Direction)
                axis.normalize()
                if fathershape:
                    size = (ArchCommands.projectToVector(fathershape.copy(), axis)).Length
                else:
                    size = 1
        if hasattr(obj, "Distance"):
            if obj.Distance.Value:
                size = obj.Distance.Value
        spacinglist = None
        if hasattr(obj, "CustomSpacing"):
            if obj.CustomSpacing:
                spacinglist = strprocessOfCustomSpacing(obj.CustomSpacing)
                influenceArea = sum(spacinglist) - spacinglist[0] / 2 - spacinglist[-1] / 2
        # Drop this check to solve issue as discussed here: https://github.com/FreeCAD/FreeCAD/pull/2550
        # if (obj.OffsetStart.Value + obj.OffsetEnd.Value) > size:
        #        return
        # all tests ok!
        if hasattr(obj, "Length"):
            length = getLengthOfRebar(obj)
            if length:
                obj.Length = length
        pl = obj.Placement
        circle = Part.makeCircle(obj.Diameter.Value / 2, bpoint, bvec)
        circle = Part.Wire(circle)
        try:
            bar = wire.makePipeShell([circle], True, False, 2)
            basewire = wire.copy()
        except Part.OCCError:
            print("Arch: error sweeping rebar profile along the base sketch")
            return
        # building final shape
        shapes = []
        placementlist = []
        self.wires = []
        rot = FreeCAD.Rotation()
        if obj.Amount == 1:
            if hasattr(obj, "RebarShape"):
                barplacement = CalculatePlacement(
                    obj.Amount,
                    1,
                    obj.Diameter.Value,
                    size,
                    axis,
                    rot,
                    obj.OffsetStart.Value,
                    obj.OffsetEnd.Value,
                    obj.RebarShape,
                )
            else:
                barplacement = CalculatePlacement(
                    obj.Amount,
                    1,
                    obj.Diameter.Value,
                    size,
                    axis,
                    rot,
                    obj.OffsetStart.Value,
                    obj.OffsetEnd.Value,
                )
            placementlist.append(barplacement)
            if hasattr(obj, "Spacing"):
                obj.Spacing = 0
        else:
            if obj.OffsetStart.Value:
                baseoffset = DraftVecUtils.scaleTo(axis, obj.OffsetStart.Value)
            else:
                baseoffset = None
            if hasattr(obj, "RebarShape") and obj.RebarShape == "Stirrup":
                interval = size - (obj.OffsetStart.Value + obj.OffsetEnd.Value + obj.Diameter.Value)
            else:
                interval = size - (obj.OffsetStart.Value + obj.OffsetEnd.Value)
            interval = interval / (obj.Amount - 1)
            for i in range(obj.Amount):
                if hasattr(obj, "RebarShape"):
                    barplacement = CalculatePlacement(
                        obj.Amount,
                        i + 1,
                        obj.Diameter.Value,
                        size,
                        axis,
                        rot,
                        obj.OffsetStart.Value,
                        obj.OffsetEnd.Value,
                        obj.RebarShape,
                    )
                else:
                    barplacement = CalculatePlacement(
                        obj.Amount,
                        i + 1,
                        obj.Diameter.Value,
                        size,
                        axis,
                        rot,
                        obj.OffsetStart.Value,
                        obj.OffsetEnd.Value,
                    )
                placementlist.append(barplacement)
            if hasattr(obj, "Spacing"):
                obj.Spacing = interval
        # Calculate placement of bars from custom spacing.
        if spacinglist:
            placementlist[:] = []
            if hasattr(obj, "RebarShape") and obj.RebarShape == "Stirrup":
                reqInfluenceArea = size - (
                    obj.OffsetStart.Value + obj.OffsetEnd.Value + obj.Diameter.Value
                )
            else:
                reqInfluenceArea = size - (obj.OffsetStart.Value + obj.OffsetEnd.Value)
            # Avoid unnecessary checks to pass like. For eg.: when we have values
            # like influenceArea is 100.00001 and reqInflueneArea is 100
            if round(influenceArea) > round(reqInfluenceArea):
                FreeCAD.Console.PrintWarning(
                    "Influence area of rebars is greater than " + str(reqInfluenceArea) + ".\n"
                )
            elif round(influenceArea) < round(reqInfluenceArea):
                FreeCAD.Console.PrintWarning("Last span is greater that end offset.\n")
            for i in range(len(spacinglist)):
                barplacement = CustomSpacingPlacement(
                    spacinglist,
                    i + 1,
                    axis,
                    father.Placement.Rotation,
                    obj.OffsetStart.Value,
                    obj.OffsetEnd.Value,
                )
                placementlist.append(barplacement)
            obj.Amount = len(spacinglist)
            obj.Spacing = 0
        obj.PlacementList = placementlist
        for i in range(len(obj.PlacementList)):
            bar = bar.copy()
            bar.Placement = obj.PlacementList[i]
            shapes.append(bar)
            w = basewire.copy()
            w.Placement = obj.PlacementList[i]
            self.wires.append(w)
        if shapes:
            obj.Shape = Part.makeCompound(shapes)
            obj.Placement = pl
        obj.TotalLength = obj.Length * len(obj.PlacementList)


class _ViewProviderRebar(ArchComponent.ViewProviderComponent):
    "A View Provider for the Rebar object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)
        self.setProperties(vobj)
        vobj.ShapeColor = ArchCommands.getDefaultColor("Rebar")

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        if not "RebarShape" in pl:
            vobj.addProperty(
                "App::PropertyString",
                "RebarShape",
                "Rebar",
                QT_TRANSLATE_NOOP("App::Property", "Shape of rebar"),
                locked=True,
            ).RebarShape
            vobj.setEditorMode("RebarShape", 2)

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Rebar_Tree.svg"

    def setEdit(self, vobj, mode):
        # The Reinforcement Workbench does not implement resetEdit.
        # Therefore unsetEdit is never called and objects would stay in
        # edit mode if this function were to return `True`.

        if mode != 0:
            return None

        if hasattr(vobj.Object, "RebarShape"):
            try:
                # Import module of RebarShape
                module = __import__(vobj.Object.RebarShape)
            except ImportError:
                FreeCAD.Console.PrintError("Unable to import RebarShape module\n")
                return False
        elif vobj.RebarShape:
            try:
                # Import module of RebarShape
                module = __import__(vobj.RebarShape)
            except ImportError:
                FreeCAD.Console.PrintError("Unable to import RebarShape module\n")
                return False
        else:
            return False

        module.editDialog(vobj)
        return False

    def updateData(self, obj, prop):

        if prop == "Shape":
            if hasattr(self, "centerline"):
                if self.centerline:
                    self.centerlinegroup.removeChild(self.centerline)
            if hasattr(obj.Proxy, "wires"):
                if obj.Proxy.wires:
                    import re
                    from pivy import coin
                    import Part

                    self.centerline = coin.SoSeparator()
                    comp = Part.makeCompound(obj.Proxy.wires)
                    buf = re.findall(r"point \[(.*?)\]", comp.writeInventor().replace("\n", ""))
                    pts = [zip(*[iter(c.split())] * 3) for c in buf]
                    for pt in pts:
                        vlist = [[float(v[0]), float(v[1]), float(v[2])] for v in pt]
                        ps = coin.SoSeparator()
                        coords = coin.SoCoordinate3()
                        coords.point.setValues(vlist)
                        ps.addChild(coords)
                        ls = coin.SoLineSet()
                        ls.numVertices = -1
                        ps.addChild(ls)
                        self.centerline.addChild(ps)
                    self.centerlinegroup.addChild(self.centerline)
        ArchComponent.ViewProviderComponent.updateData(self, obj, prop)

    def attach(self, vobj):

        from pivy import coin

        self.centerlinegroup = coin.SoSeparator()
        self.centerlinegroup.setName("Centerline")
        self.centerlinecolor = coin.SoBaseColor()
        self.centerlinestyle = coin.SoDrawStyle()
        self.centerlinegroup.addChild(self.centerlinecolor)
        self.centerlinegroup.addChild(self.centerlinestyle)
        vobj.addDisplayMode(self.centerlinegroup, "Centerline")
        ArchComponent.ViewProviderComponent.attach(self, vobj)

    def onChanged(self, vobj, prop):

        if (prop == "LineColor") and hasattr(vobj, "LineColor"):
            if hasattr(self, "centerlinecolor"):
                c = vobj.LineColor
                self.centerlinecolor.rgb.setValue(c[0], c[1], c[2])
        elif (prop == "LineWidth") and hasattr(vobj, "LineWidth"):
            if hasattr(self, "centerlinestyle"):
                self.centerlinestyle.lineWidth = vobj.LineWidth
        ArchComponent.ViewProviderComponent.onChanged(self, vobj, prop)

    def getDisplayModes(self, vobj):

        modes = ["Centerline"]
        return modes + ArchComponent.ViewProviderComponent.getDisplayModes(self, vobj)


def CalculatePlacement(
    baramount, barnumber, bardiameter, size, axis, rotation, offsetstart, offsetend, RebarShape=""
):
    """CalculatePlacement([baramount, barnumber, bardiameter, size, axis, rotation, offsetstart, offsetend, RebarShape]):
    Calculate the placement of the bar from given values."""
    if baramount == 1:
        interval = offsetstart
    else:
        if RebarShape == "Stirrup":
            interval = size - (offsetstart + offsetend + bardiameter)
        else:
            interval = size - (offsetstart + offsetend)
        interval = interval / (baramount - 1)
    bardistance = (interval * (barnumber - 1)) + offsetstart
    barplacement = DraftVecUtils.scaleTo(axis, bardistance)
    placement = FreeCAD.Placement(barplacement, rotation)
    return placement


def CustomSpacingPlacement(spacinglist, barnumber, axis, rotation, offsetstart, offsetend):
    """CustomSpacingPlacement(spacinglist, barnumber, axis, rotation, offsetstart, offsetend):
    Calculate placement of the bar from custom spacing list."""
    if barnumber == 1:
        bardistance = offsetstart
    else:
        bardistance = sum(spacinglist[0:barnumber])
        bardistance = bardistance - spacinglist[0] / 2
        bardistance = bardistance - spacinglist[barnumber - 1] / 2
        bardistance = bardistance + offsetstart
    barplacement = DraftVecUtils.scaleTo(axis, bardistance)
    placement = FreeCAD.Placement(barplacement, rotation)
    return placement


def strprocessOfCustomSpacing(span_string):
    """strprocessOfCustomSpacing(span_string): This function take input
    in specific syntax and return output in the form of list. For eg.
    Input: "3@100+2@200+3@100"
    Output: [100, 100, 100, 200, 200, 100, 100, 100]"""
    # import string
    span_st = span_string.strip()
    span_sp = span_st.split("+")
    index = 0
    spacinglist = []
    while index < len(span_sp):
        # Find "@" recursively in span_sp array.
        # If not found, append the index value to "spacinglist" array.
        if span_sp[index].find("@") == -1:
            spacinglist.append(float(span_sp[index]))
        else:
            in_sp = span_sp[index].split("@")
            count = 0
            while count < int(in_sp[0]):
                spacinglist.append(float(in_sp[1]))
                count += 1
        index += 1
    return spacinglist


def getLengthOfRebar(rebar):
    """getLengthOfRebar(RebarObject): Calculates the length of the rebar."""
    base = rebar.Base
    # When rebar is derived from DWire
    if hasattr(base, "Length"):
        return base.Length
    # When rebar is derived from Sketch
    elif base.isDerivedFrom("Sketcher::SketchObject"):
        length = 0
        for geo in base.Geometry:
            length += geo.length()
        return length
    elif base.isDerivedFrom("Part::Helix"):
        return base.Shape.Wires[0].Length
    else:
        FreeCAD.Console.PrintError("Cannot calculate rebar length from its base object\n")
        return None



# ***************************************************************************
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Tools for FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import math
import os
import sys
import tempfile
import unittest
from typing import Union, List, Iterator

import FreeCAD

from os.path import join


def get_fem_test_home_dir():
    return join(FreeCAD.getHomePath(), "Mod", "Fem", "femtest", "data")


def get_fem_test_tmp_dir(dirname=None):
    from uuid import uuid4

    _unique_id = str(uuid4())[-12:]
    # print(_unique_id)
    if dirname is None:
        temp_dir = join(tempfile.gettempdir(), "FEM_unittests", _unique_id)
    else:
        temp_dir = join(tempfile.gettempdir(), "FEM_unittests", dirname + "_" + _unique_id)
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)
    return temp_dir


def get_unit_test_tmp_dir(temp_dir, unittestdir):
    testdir = join(temp_dir, unittestdir)
    if not os.path.exists(testdir):
        os.makedirs(testdir)
    return testdir


def fcc_print(message):
    FreeCAD.Console.PrintMessage(f"{message} \n")


def get_namefromdef(strdel="", stradd=""):
    # https://code.activestate.com/recipes/66062-determining-current-function-name/
    return (sys._getframe(1).f_code.co_name).replace(strdel, stradd)


def get_defmake_count(fem_vtk_post=True):
    """
    count the def make in module ObjectsFem
    could also be done in bash with
    grep -c  "def make" src/Mod/Fem/ObjectsFem.py
    """
    name_modfile = join(FreeCAD.getHomePath(), "Mod", "Fem", "ObjectsFem.py")
    modfile = open(name_modfile)
    lines_modefile = modfile.readlines()
    modfile.close()
    lines_defmake = [li for li in lines_modefile if li.startswith("def make")]
    if not fem_vtk_post:
        # FEM VTK post processing is disabled
        # we are not able to create VTK post objects
        new_lines = []
        for li in lines_defmake:
            if "Post" not in li:
                new_lines.append(li)
        lines_defmake = new_lines
    return len(lines_defmake)


def get_fem_test_defs():

    test_path = join(FreeCAD.getHomePath(), "Mod", "Fem", "femtest", "app")
    print(f"Modules, classes, methods taken from: {test_path}")

    collected_test_module_paths = []
    for tfile in sorted(os.listdir(test_path)):
        if tfile.startswith("test") and tfile.endswith(".py"):
            collected_test_module_paths.append(join(test_path, tfile))

    collected_test_modules = []
    collected_test_classes = []
    collected_test_methods = []
    for f in collected_test_module_paths:
        module_name = os.path.splitext(os.path.basename(f))[0]
        module_path = f"femtest.app.{module_name}"
        if module_path not in collected_test_modules:
            collected_test_modules.append(module_path)
        class_name = ""
        tfile = open(f)
        for ln in tfile:
            ln = ln.lstrip()
            ln = ln.rstrip()
            if ln.startswith("class "):
                ln = ln.lstrip("class ")
                ln = ln.split("(")[0]
                class_name = ln
                class_path = f"femtest.app.{module_name}.{class_name}"
                if class_path not in collected_test_classes:
                    collected_test_classes.append(class_path)
            if ln.startswith("def test"):
                ln = ln.lstrip("def ")
                ln = ln.split("(")[0]
                if ln == "test_00print":
                    continue
                method_path = f"femtest.app.{module_name}.{class_name}.{ln}"
                collected_test_methods.append(method_path)
        tfile.close()

    # write to file
    file_path = join(tempfile.gettempdir(), "test_commands.sh")
    cf = open(file_path, "w")
    cf.write("# created by Python\n")
    cf.write("'''\n")
    cf.write("from femtest.app.support_utils import get_fem_test_defs\n")
    cf.write("get_fem_test_defs()\n")
    cf.write("\n")
    cf.write("\n")
    cf.write("# all FEM App tests\n")
    cf.write("make -j 4 && ./bin/FreeCAD --run-test 'TestFemApp'\n")
    cf.write("\n")
    cf.write("make -j 4 && ./bin/FreeCADCmd --run-test 'TestFemApp'\n")
    cf.write("\n")
    cf.write("\n")
    cf.write("'''\n")
    cf.write("\n")
    cf.write("# modules\n")
    for m in collected_test_modules:
        cf.write(f"make -j 4 && ./bin/FreeCADCmd -t {m}\n")
    cf.write("\n")
    cf.write("\n")
    cf.write("# classes\n")
    for m in collected_test_classes:
        cf.write(f"make -j 4 && ./bin/FreeCADCmd -t {m}\n")
    cf.write("\n")
    cf.write("\n")
    cf.write("# methods\n")
    for m in collected_test_methods:
        cf.write(f"make -j 4 && ./bin/FreeCADCmd -t {m}\n")
    cf.write("\n")
    cf.write("\n")
    cf.write("# methods in FreeCAD\n")
    for m in collected_test_methods:
        cf.write(
            "\nimport unittest\n"
            "unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(\n"
            "    '{}'\n"
            "))\n".format(m)
        )
    cf.close()
    print(f"The file was saved in:{file_path}")


def try_converting_to_float(line: str) -> Union[None, List[float]]:
    """Does its best to split a line and convert its elements to float

    Has 3 strategies of splitting:
        * by comma - mainly in CalculiX .inp files
        * by space - other solvers
        * by space and ignoring the 1st word - other solvers
            If there was only 1 word, the line will fail, no compromises

    Single characters always pass

    :param line: line to split and convert to floats
    :return: None if conversion failed, else list of floats
    """
    strategies = [
        lambda _line: _line[1:].split(","),
        lambda _line: _line[1:].split(),
        lambda _line: (lambda split_line: len(split_line) > 1 and split_line or "fail")(
            _line[1:].split()[1:]
        ),
    ]
    for strategy in strategies:
        try:
            return list(map(float, strategy(line)))
        except ValueError:
            pass
    return None


def are_floats_equal(
    orig_floats: Union[None, List[float]], floats_to_compare: Union[None, List[float]]
) -> bool:
    """Check if floats in lists are equal with some tolerance

    :param orig_floats: list of floats - left operands of comparison
    :param floats_to_compare: list of floats - right operands of comparison
    :return: True if both lists are equal with some tolerance element-wise, else False
    """
    if any(floats is None for floats in (orig_floats, floats_to_compare)):
        return False
    for orig_float, float_to_compare in zip(orig_floats, floats_to_compare):
        if not math.isclose(orig_float, float_to_compare, abs_tol=1e-10):
            return False
    return True


def parse_diff(diff_lines: Iterator[str]) -> List[str]:
    """Parses lines from `united_diff`

    Tries to split a line and convert its contents to float, to
    compare float numbers with some tolerance

    Recognizes blocks of changes, that start with `@@` and end either at EOF or
    at the beginning of another block. Only bad lines are added to block.
    All lines, which didn't pass the element-wise check, are bad lines.
    :param diff_lines: lines produced by `united_diff`
    :return: list of bad lines with respect to their block of change
    """
    bad_lines = []
    changes_block = []
    while True:
        try:
            first = next(diff_lines)
            if first.startswith("@@"):
                if len(changes_block) > 1:
                    bad_lines.extend(changes_block)
                changes_block = [first]
                continue
            second = next(diff_lines)
        except StopIteration:
            break
        if first.startswith("---"):
            continue
        if not are_floats_equal(*map(try_converting_to_float, (first, second))):
            changes_block.extend((first, second))
    # check if we have any remaining block
    if len(changes_block) > 1:
        bad_lines.extend(changes_block)
    return bad_lines


def compare_inp_files(file_name1, file_name2):
    file1 = open(file_name1)
    f1 = file1.readlines()
    file1.close()
    # l.startswith("17671.0,1") is a temporary workaround
    # for python3 problem with 1DFlow input
    # TODO as soon as the 1DFlow result reading is fixed
    # this should be triggered in the 1DFlow unit test
    lf1 = [
        li
        for li in f1
        if not (
            li.startswith("**   written ")
            or li.startswith("**   file ")
            or li.startswith("17671.0,1")
        )
    ]
    lf1 = force_unix_line_ends(lf1)
    file2 = open(file_name2)
    f2 = file2.readlines()
    file2.close()
    # TODO see comment on file1
    lf2 = [
        li
        for li in f2
        if not (
            li.startswith("**   written ")
            or li.startswith("**   file ")
            or li.startswith("17671.0,1")
        )
    ]
    lf2 = force_unix_line_ends(lf2)
    import difflib

    diff_lines = difflib.unified_diff(lf1, lf2, n=0)

    bad_lines = parse_diff(diff_lines)
    if bad_lines:
        return f"Comparing {file_name1} to {file_name2} failed!\n{''.join(bad_lines)}"


def compare_files(file_name1, file_name2):
    file1 = open(file_name1)
    f1 = file1.readlines()
    file1.close()

    # TODO: add support for variable values in the reference file
    # instead of using this workaround

    # workaround to compare geos of elmer test and temporary file path
    # (not only names change, path changes with operating system)
    lf1 = [
        li
        for li in f1
        if not (
            li.startswith('Merge "')
            or li.startswith('Save "')
            or li.startswith("// ")
            or li.startswith("General.NumThreads")
        )
    ]
    lf1 = force_unix_line_ends(lf1)
    file2 = open(file_name2)
    f2 = file2.readlines()
    file2.close()
    lf2 = [
        li
        for li in f2
        if not (
            li.startswith('Merge "')
            or li.startswith('Save "')
            or li.startswith("// ")
            or li.startswith("General.NumThreads")
        )
    ]
    lf2 = force_unix_line_ends(lf2)
    import difflib

    diff = difflib.unified_diff(lf1, lf2, n=0)
    result = ""
    for li in diff:
        result += li
    if result:
        result = f"Comparing {file_name1} to {file_name2} failed!\n" + result
    return result


def compare_stats(fea, stat_file, res_obj_name, loc_stat_types=None):
    import femresult.resulttools as resulttools

    # get the stat types which should be compared
    stat_types = [
        "U1",
        "U2",
        "U3",
        "Uabs",
        "Sabs",
        "MaxPrin",
        "MidPrin",
        "MinPrin",
        "MaxShear",
        "Peeq",
        "Temp",
        "MFlow",
        "NPress",
    ]
    if not loc_stat_types:
        loc_stat_types = stat_types

    # get stats from result obj which should be compared
    obj = fea.analysis.Document.getObject(res_obj_name)
    # fcc_print(obj)
    if obj:
        # fcc_print(obj.Name)
        stats = []
        for s in loc_stat_types:
            statval = resulttools.get_stats(obj, s)
            stats.append(f"{s}: ({statval[0]:.10f}, {statval[1]:.10f})\n")
    else:
        fcc_print(f"Result object not found. Name: {res_obj_name}")
        return True

    # get stats to compare with, the expected ones
    sf = open(stat_file)
    sf_content = []
    for li in sf.readlines():
        for st in loc_stat_types:
            if li.startswith(st):
                sf_content.append(li)
    sf.close()
    sf_content = force_unix_line_ends(sf_content)
    if sf_content == []:
        return True

    # compare stats
    if stats != sf_content:
        fcc_print(f"Stats read from {fea.base_name}.frd file")
        fcc_print("!=")
        fcc_print(f"Expected stats from {stat_file}")
        for i in range(len(stats)):
            if stats[i] != sf_content[i]:
                fcc_print(f"{stats[i].rstrip()} != {sf_content[i].rstrip()}")
        return True

    return False


def force_unix_line_ends(line_list):
    new_line_list = []
    for ln in line_list:
        if ln.endswith("\r\n"):
            ln = ln[:-2] + "\n"
        new_line_list.append(ln)
    return new_line_list


def collect_python_modules(femsubdir=None):
    if not femsubdir:
        pydir = join(FreeCAD.ConfigGet("AppHomePath"), "Mod", "Fem")
    else:
        pydir = join(FreeCAD.ConfigGet("AppHomePath"), "Mod", "Fem", femsubdir)
    collected_modules = []
    fcc_print(pydir)
    for pyfile in sorted(os.listdir(pydir)):
        if pyfile.endswith(".py") and not pyfile.startswith("Init"):
            if not femsubdir:
                collected_modules.append(os.path.splitext(os.path.basename(pyfile))[0])
            else:
                collected_modules.append(
                    femsubdir.replace("/", ".")
                    + "."
                    + os.path.splitext(os.path.basename(pyfile))[0]
                )
    return collected_modules


def all_test_files():
    # open all files
    cube_frequency()
    cube_static()
    Flow1D_thermomech()
    multimat()
    spine_thermomech()


# run the specific test case of the file
# open the file in FreeCAD GUI and return the doc identifier
def cube_frequency():
    testname = "femtest.testccxtools.TestCcxTools.test_3_freq_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(get_fem_test_tmp_dir(), "FEM_ccx_frequency", "cube_frequency.FCStd"))
    return doc


def cube_static():
    testname = "femtest.testccxtools.TestCcxTools.test_1_static_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(get_fem_test_tmp_dir(), "FEM_ccx_static", "cube_static.FCStd"))
    return doc


def Flow1D_thermomech():
    testname = "femtest.testccxtools.TestCcxTools.test_5_Flow1D_thermomech_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(
        join(get_fem_test_tmp_dir(), "FEM_ccx_Flow1D_thermomech", "Flow1D_thermomech.FCStd")
    )
    return doc


def multimat():
    testname = "femtest.testccxtools.TestCcxTools.test_2_static_multiple_material"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(get_fem_test_tmp_dir(), "FEM_ccx_multimat", "multimat.FCStd"))
    return doc


def spine_thermomech():
    testname = "femtest.testccxtools.TestCcxTools.test_4_thermomech_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(get_fem_test_tmp_dir(), "FEM_ccx_thermomech", "spine_thermomech.FCStd"))
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *   Copyright (c) 2018, 2019 Gauthier Briere                              *
# *   Copyright (c) 2019, 2020 Schildkroet                                  *
# *   Copyright (c) 2022-2025 Larry Woestman <LarryWoestman2@gmail.com>     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import datetime
import os
from typing import Any, Dict, List

import FreeCAD
import Path.Base.Util as PathUtil
import Path.Post.Utils as PostUtils
import Path.Post.UtilsParse as PostUtilsParse
import Path.Tool.Controller as PathToolController

# Define some types that are used throughout this file
Gcode = List[str]
Values = Dict[str, Any]


def check_canned_cycles(values: Values) -> None:
    """Check canned cycles for drilling."""
    if values["TRANSLATE_DRILL_CYCLES"]:
        if len(values["SUPPRESS_COMMANDS"]) == 0:
            values["SUPPRESS_COMMANDS"] = ["G99", "G98", "G80"]
        else:
            values["SUPPRESS_COMMANDS"] += ["G99", "G98", "G80"]


def output_coolant_off(values: Values, gcode: Gcode, coolant_mode: str) -> None:
    """Output the commands to turn coolant off if necessary."""
    comment: str

    if values["ENABLE_COOLANT"] and coolant_mode != "None":
        if values["OUTPUT_COMMENTS"]:
            comment = PostUtilsParse.create_comment(values, f"Coolant Off: {coolant_mode}")
            gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        gcode.append(f"{PostUtilsParse.linenumber(values)}M9")


def output_coolant_on(values: Values, gcode: Gcode, coolant_mode: str) -> None:
    """Output the commands to turn coolant on if necessary."""
    comment: str

    if values["ENABLE_COOLANT"]:
        if values["OUTPUT_COMMENTS"] and coolant_mode != "None":
            comment = PostUtilsParse.create_comment(values, f"Coolant On: {coolant_mode}")
            gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        if coolant_mode == "Flood":
            gcode.append(f"{PostUtilsParse.linenumber(values)}M8")
        elif coolant_mode == "Mist":
            gcode.append(f"{PostUtilsParse.linenumber(values)}M7")


def output_end_bcnc(values: Values, gcode: Gcode) -> None:
    """Output the ending BCNC header."""
    comment: str

    if values["OUTPUT_BCNC"]:
        comment = PostUtilsParse.create_comment(values, "Block-name: post_amble")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        comment = PostUtilsParse.create_comment(values, "Block-expand: 0")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        comment = PostUtilsParse.create_comment(values, "Block-enable: 1")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")


def output_header(values: Values, gcode: Gcode) -> None:
    """Output the header."""
    cam_file: str
    comment: str

    if not values["OUTPUT_HEADER"]:
        return
    comment = PostUtilsParse.create_comment(values, "Exported by FreeCAD")
    gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
    comment = PostUtilsParse.create_comment(
        values, f'Post Processor: {values["POSTPROCESSOR_FILE_NAME"]}'
    )
    gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
    if FreeCAD.ActiveDocument:
        cam_file = os.path.basename(FreeCAD.ActiveDocument.FileName)
    else:
        cam_file = "<None>"
    comment = PostUtilsParse.create_comment(values, f"Cam File: {cam_file}")
    gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
    comment = PostUtilsParse.create_comment(values, f"Output Time: {str(datetime.datetime.now())}")
    gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")


def output_motion_mode(values: Values, gcode: Gcode) -> None:
    """Verify if PREAMBLE or SAFETYBLOCK have changed MOTION_MODE."""

    if "G90" in values["PREAMBLE"] or "G90" in values["SAFETYBLOCK"]:
        values["MOTION_MODE"] = "G90"
    elif "G91" in values["PREAMBLE"] or "G91" in values["SAFETYBLOCK"]:
        values["MOTION_MODE"] = "G91"
    else:
        gcode.append(f'{PostUtilsParse.linenumber(values)}{values["MOTION_MODE"]}')


def output_postamble_header(values: Values, gcode: Gcode) -> None:
    """Output the postamble header."""
    comment: str = ""

    if values["OUTPUT_COMMENTS"]:
        comment = PostUtilsParse.create_comment(values, "Begin postamble")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")


def output_postamble(values: Values, gcode: Gcode) -> None:
    """Output the postamble."""
    line: str

    for line in values["POSTAMBLE"].splitlines(False):
        gcode.append(f"{PostUtilsParse.linenumber(values)}{line}")


def output_postop(values: Values, gcode: Gcode, obj) -> None:
    """Output the post-operation information."""
    comment: str
    line: str

    if values["OUTPUT_COMMENTS"]:
        if values["SHOW_OPERATION_LABELS"]:
            comment = PostUtilsParse.create_comment(
                values, f'{values["FINISH_LABEL"]} operation: {obj.Label}'
            )
        else:
            comment = PostUtilsParse.create_comment(values, f'{values["FINISH_LABEL"]} operation')
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
    for line in values["POST_OPERATION"].splitlines(False):
        gcode.append(f"{PostUtilsParse.linenumber(values)}{line}")


def output_preamble(values: Values, gcode: Gcode) -> None:
    """Output the preamble."""
    comment: str
    line: str

    if values["OUTPUT_COMMENTS"]:
        comment = PostUtilsParse.create_comment(values, "Begin preamble")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
    for line in values["PREAMBLE"].splitlines(False):
        gcode.append(f"{PostUtilsParse.linenumber(values)}{line}")


def output_preop(values: Values, gcode: Gcode, obj) -> None:
    """Output the pre-operation information."""
    comment: str
    line: str

    if values["OUTPUT_COMMENTS"]:
        if values["SHOW_OPERATION_LABELS"]:
            comment = PostUtilsParse.create_comment(values, f"Begin operation: {obj.Label}")
        else:
            comment = PostUtilsParse.create_comment(values, "Begin operation")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        if values["SHOW_MACHINE_UNITS"]:
            comment = PostUtilsParse.create_comment(
                values, f'Machine units: {values["UNIT_SPEED_FORMAT"]}'
            )
            gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        if values["OUTPUT_MACHINE_NAME"]:
            comment = PostUtilsParse.create_comment(
                values,
                f'Machine: {values["MACHINE_NAME"]}, {values["UNIT_SPEED_FORMAT"]}',
            )
            gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
    for line in values["PRE_OPERATION"].splitlines(False):
        gcode.append(f"{PostUtilsParse.linenumber(values)}{line}")


def output_return_to(values: Values, gcode: Gcode) -> None:
    """Output the RETURN_TO command."""
    cmd: str
    num_x: str
    num_y: str
    num_z: str

    if values["RETURN_TO"]:
        num_x = values["RETURN_TO"][0]
        num_y = values["RETURN_TO"][1]
        num_z = values["RETURN_TO"][2]
        cmd = PostUtilsParse.format_command_line(
            values, ["G0", f"X{num_x}", f"Y{num_y}", f"Z{num_z}"]
        )
        gcode.append(f"{PostUtilsParse.linenumber(values)}{cmd}")


def output_safetyblock(values: Values, gcode: Gcode) -> None:
    """Output the safety block."""
    line: str

    for line in values["SAFETYBLOCK"].splitlines(False):
        gcode.append(f"{PostUtilsParse.linenumber(values)}{line}")


def output_start_bcnc(values: Values, gcode: Gcode, obj) -> None:
    """Output the starting BCNC header."""
    comment: str

    if values["OUTPUT_BCNC"]:
        comment = PostUtilsParse.create_comment(values, f"Block-name: {obj.Label}")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        comment = PostUtilsParse.create_comment(values, "Block-expand: 0")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")
        comment = PostUtilsParse.create_comment(values, "Block-enable: 1")
        gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")


def output_tool_list(values: Values, gcode: Gcode, objectslist) -> None:
    """Output a list of the tools used in the objects."""
    comment: str

    if values["OUTPUT_COMMENTS"] and values["LIST_TOOLS_IN_PREAMBLE"]:
        for item in objectslist:
            if hasattr(item, "Proxy") and isinstance(item.Proxy, PathToolController.ToolController):
                comment = PostUtilsParse.create_comment(values, f"T{item.ToolNumber}={item.Name}")
                gcode.append(f"{PostUtilsParse.linenumber(values)}{comment}")


def output_tool_return(values: Values, gcode: Gcode) -> None:
    """Output the tool return block."""
    line: str

    for line in values["TOOLRETURN"].splitlines(False):
        gcode.append(f"{PostUtilsParse.linenumber(values)}{line}")


def output_units(values: Values, gcode: Gcode) -> None:
    """Verify if PREAMBLE or SAFETYBLOCK have changed UNITS."""

    if "G21" in values["PREAMBLE"] or "G21" in values["SAFETYBLOCK"]:
        values["UNITS"] = "G21"
        values["UNIT_FORMAT"] = "mm"
        values["UNIT_SPEED_FORMAT"] = "mm/min"
    elif "G20" in values["PREAMBLE"] or "G20" in values["SAFETYBLOCK"]:
        values["UNITS"] = "G20"
        values["UNIT_FORMAT"] = "in"
        values["UNIT_SPEED_FORMAT"] = "in/min"
    else:
        gcode.append(f'{PostUtilsParse.linenumber(values)}{values["UNITS"]}')


def export_common(values: Values, objectslist, filename: str) -> str:
    """Do the common parts of postprocessing the objects in objectslist to filename."""
    coolant_mode: str
    dia: PostUtils.GCodeEditorDialog
    final: str
    final_for_editor: str
    gcode: Gcode = []

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(f"The object {obj.Name} is not a path.")
            print("Please select only path and Compounds.")
            return ""

    print(f'PostProcessor:  {values["POSTPROCESSOR_FILE_NAME"]} postprocessing...')

    check_canned_cycles(values)
    output_header(values, gcode)
    output_safetyblock(values, gcode)
    output_tool_list(values, gcode, objectslist)
    output_preamble(values, gcode)
    output_motion_mode(values, gcode)
    output_units(values, gcode)

    for obj in objectslist:
        # Skip inactive operations
        if not PathUtil.activeForOp(obj):
            continue
        coolant_mode = PathUtil.coolantModeForOp(obj)
        output_start_bcnc(values, gcode, obj)
        output_preop(values, gcode, obj)
        output_coolant_on(values, gcode, coolant_mode)
        # output the G-code for the group (compound) or simple path
        PostUtilsParse.parse_a_group(values, gcode, obj)
        output_postop(values, gcode, obj)
        output_coolant_off(values, gcode, coolant_mode)

    output_return_to(values, gcode)
    #
    # This doesn't make sense to me.  It seems that both output_start_bcnc and
    # output_end_bcnc should be in the for loop or both should be out of the
    # for loop.  However, that is the way that grbl post code was written, so
    # for now I will leave it that way until someone has time to figure it out.
    #
    output_end_bcnc(values, gcode)
    output_postamble_header(values, gcode)
    output_tool_return(values, gcode)
    output_safetyblock(values, gcode)
    output_postamble(values, gcode)

    # add the appropriate end-of-line characters to the gcode, including after the last line
    gcode.append("")
    if values["END_OF_LINE_CHARACTERS"] == "\n\n":
        # flag that we want to use "\n" as the end-of-line characters
        # by putting "\n\n" at the front of the gcode (which shouldn't otherwise happen)
        final = "\n\n" + "\n".join(gcode)
    else:
        # the other possibilities are:
        #    "\n"   means "use the end-of-line characters that match the system"
        #    "\r"   means "use \r"
        #    "\r\n" means "use \r\n"
        final = values["END_OF_LINE_CHARACTERS"].join(gcode)

    if FreeCAD.GuiUp and values["SHOW_EDITOR"]:
        if len(final) > 100000:
            print("Skipping editor since output is greater than 100kb")
        else:
            dia = PostUtils.GCodeEditorDialog()
            # the editor expects lines to end in "\n", and returns lines ending in "\n"
            if values["END_OF_LINE_CHARACTERS"] == "\n":
                dia.editor.setText(final)
                if dia.exec_():
                    final = dia.editor.toPlainText()
            else:
                final_for_editor = "\n".join(gcode)
                dia.editor.setText(final_for_editor)
                if dia.exec_():
                    final_for_editor = dia.editor.toPlainText()
                    # convert all "\n" to the appropriate end-of-line characters
                    if values["END_OF_LINE_CHARACTERS"] == "\n\n":
                        # flag that we want to use "\n" as the end-of-line characters
                        # by putting "\n\n" at the front of the gcode
                        # (which shouldn't otherwise happen)
                        final = "\n\n" + final_for_editor
                    else:
                        # the other possibilities are:
                        #    "\r"   means "use \r"
                        #    "\r\n" means "use \r\n"
                        final = final_for_editor.replace("\n", values["END_OF_LINE_CHARACTERS"])

    print("done postprocessing.")

    if not filename == "-":
        if final[0:2] == "\n\n":
            # write out the gcode using "\n" as the end-of-line characters
            with open(filename, "w", encoding="utf-8", newline="") as gfile:
                gfile.write(final[2:])
        elif "\r" in final:
            with open(filename, "w", encoding="utf-8", newline="") as gfile:
                # write out the gcode with whatever end-of-line characters it already has,
                # presumably either "\r" or "\r\n"
                gfile.write(final)
        else:
            with open(filename, "w", encoding="utf-8", newline=None) as gfile:
                # The gcode has "\n" as the end-of-line characters, which means
                # "write out the gcode with whatever end-of-line characters the system
                # that is running the postprocessor uses".
                gfile.write(final)

    return final



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD task panel base for post object task panels"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package base_fempostpanel
#  \ingroup FEM
#  \brief task panel base for post objects

from PySide import QtCore, QtGui

import FreeCAD

from . import base_femtaskpanel

translate = FreeCAD.Qt.translate


class _BasePostTaskPanel(base_femtaskpanel._BaseTaskPanel):
    """
    The TaskPanel for post objects, mimicking the c++ functionality
    """

    def __init__(self, obj):
        super().__init__(obj)

        # get the settings group
        self.__settings_grp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem")

    # Implement parent functions
    # ##########################

    def getStandardButtons(self):
        return (
            QtGui.QDialogButtonBox.Apply | QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel
        )

    def clicked(self, button):
        # apply button hit?
        if button == QtGui.QDialogButtonBox.Apply:
            self.obj.Document.recompute()

    def open(self):
        # open a new transaction if non is open
        if not FreeCAD.getActiveTransaction():
            FreeCAD.ActiveDocument.openTransaction(
                translate("FEM", "Edit {}").format(self.obj.Label)
            )

    # Helper functions
    # ################

    def _recompute(self):
        # only recompute if the user wants automatic recompute
        if self.__settings_grp.GetBool("PostAutoRecompute", True):
            self.obj.Document.recompute()

    def _enumPropertyToCombobox(self, obj, prop, cbox):
        cbox.blockSignals(True)
        cbox.clear()
        entries = obj.getEnumerationsOfProperty(prop)
        for entry in entries:
            cbox.addItem(entry)

        cbox.setCurrentText(getattr(obj, prop))
        cbox.blockSignals(False)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

## @package ArchSectionPlane
#  \ingroup ARCH
#  \brief The Section plane object and tools
#
#  This module provides tools to build Section plane objects.
#  It also contains functionality to produce SVG rendering of
#  section planes, to be used in the TechDraw module

import math
import re
import tempfile
import time
import uuid

import FreeCAD
import ArchCommands
import ArchComponent
import Draft
import DraftVecUtils

from FreeCAD import Vector
from draftutils import gui_utils
from draftutils import params

if FreeCAD.GuiUp:
    from pivy import coin
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond

ISRENDERING = False  # flag to prevent concurrent runs of the coin renderer


def getSectionData(source):
    """Returns some common data from section planes and building parts"""

    if hasattr(source, "Objects"):
        objs = source.Objects
        cutplane = source.Shape
    elif hasattr(source, "Group"):
        import Part

        objs = source.Group
        cutplane = Part.makePlane(1000, 1000, FreeCAD.Vector(-500, -500, 0))
        m = 1
        if source.ViewObject and hasattr(source.ViewObject, "CutMargin"):
            m = source.ViewObject.CutMargin.Value
        cutplane.translate(FreeCAD.Vector(0, 0, m))
        cutplane.Placement = cutplane.Placement.multiply(source.Placement)
    onlySolids = True
    if hasattr(source, "OnlySolids"):
        onlySolids = source.OnlySolids
    clip = False
    if hasattr(source, "Clip"):
        clip = source.Clip
    p = FreeCAD.Placement(source.Placement)
    direction = p.Rotation.multVec(FreeCAD.Vector(0, 0, 1))
    if objs:
        objs = Draft.get_group_contents(objs, walls=True, addgroups=True)
    return objs, cutplane, onlySolids, clip, direction


def getCutShapes(
    objs, cutplane, onlySolids, clip, joinArch, showHidden, groupSshapesByObject=False
):
    """
    returns a list of shapes (visible, hidden, cut lines...)
    obtained from performing a series of booleans against the given cut plane
    """

    import Part
    import DraftGeomUtils

    shapes = []
    hshapes = []
    sshapes = []
    objectShapes = []
    objectSshapes = []

    if joinArch:
        shtypes = {}
        for o in objs:
            if Draft.getType(o) in ["Wall", "Structure"]:
                if o.Shape.isNull():
                    pass
                elif onlySolids:
                    shtypes.setdefault(
                        o.Material.Name if (hasattr(o, "Material") and o.Material) else "None", []
                    ).extend(o.Shape.Solids)
                else:
                    shtypes.setdefault(
                        o.Material.Name if (hasattr(o, "Material") and o.Material) else "None", []
                    ).append(o.Shape.copy())
            elif hasattr(o, "Shape"):
                if o.Shape.isNull():
                    pass
                elif onlySolids:
                    shapes.extend(o.Shape.Solids)
                    objectShapes.append((o, o.Shape.Solids))
                else:
                    shapes.append(o.Shape.copy())
                    objectShapes.append((o, [o.Shape.copy()]))
        for k, v in shtypes.items():
            v1 = v.pop()
            if v:
                v1 = v1.multiFuse(v)
                v1 = v1.removeSplitter()
            if v1.Solids:
                shapes.extend(v1.Solids)
                objectShapes.append((k, v1.Solids))
            else:
                print("ArchSectionPlane: Fusing Arch objects produced non-solid results")
                shapes.append(v1)
                objectShapes.append((k, [v1]))
    else:
        for o in objs:
            if hasattr(o, "Shape"):
                if o.Shape.isNull():
                    pass
                elif onlySolids:
                    if o.Shape.isValid():
                        shapes.extend(o.Shape.Solids)
                        objectShapes.append((o, o.Shape.Solids))
                else:
                    shapes.append(o.Shape)
                    objectShapes.append((o, [o.Shape]))

    cutface, cutvolume, invcutvolume = ArchCommands.getCutVolume(cutplane, shapes, clip)
    shapes = []
    for o, shapeList in objectShapes:
        tmpSshapes = []
        for sh in shapeList:
            for sub in (sh.SubShapes if sh.ShapeType == "Compound" else [sh]):
                if cutvolume:
                    if sub.Volume < 0:
                        sub = sub.reversed()  # Use reversed as sub is immutable.
                    c = sub.cut(cutvolume)
                    s = sub.section(cutface)
                    try:
                        wires = DraftGeomUtils.findWires(s.Edges)
                        for w in wires:
                            f = Part.Face(w)
                            tmpSshapes.append(f)
                    except Part.OCCError:
                        # print "ArchView: unable to get a face"
                        tmpSshapes.append(s)
                    shapes.extend(c.SubShapes if c.ShapeType == "Compound" else [c])
                    if showHidden:
                        c = sub.cut(invcutvolume)
                        hshapes.extend(c.SubShapes if c.ShapeType == "Compound" else [c])
                else:
                    shapes.append(sub)

            if len(tmpSshapes) > 0:
                sshapes.extend(tmpSshapes)

                if groupSshapesByObject:
                    objectSshapes.append((o, tmpSshapes))

    if groupSshapesByObject:
        return shapes, hshapes, sshapes, cutface, cutvolume, invcutvolume, objectSshapes
    else:
        return shapes, hshapes, sshapes, cutface, cutvolume, invcutvolume


def getFillForObject(o, defaultFill, source):
    """returns a color tuple from an object's material"""

    if hasattr(source, "UseMaterialColorForFill") and source.UseMaterialColorForFill:
        material = None
        if hasattr(o, "Material") and o.Material:
            material = o.Material
        elif isinstance(o, str):
            material = FreeCAD.ActiveDocument.getObject(o)
        if material:
            if hasattr(material, "SectionColor") and material.SectionColor:
                return material.SectionColor
            elif hasattr(material, "Color") and material.Color:
                return material.Color
        elif hasattr(o, "ViewObject") and hasattr(o.ViewObject, "ShapeColor"):
            return o.ViewObject.ShapeColor
    return defaultFill


def isOriented(obj, plane):
    """determines if an annotation is facing the cutplane or not"""

    norm1 = plane.normalAt(0, 0)
    if hasattr(obj, "Placement"):
        norm2 = obj.Placement.Rotation.multVec(FreeCAD.Vector(0, 0, 1))
    elif hasattr(obj, "Normal"):
        norm2 = obj.Normal
        if norm2.Length < 0.01:
            return True
    else:
        return True
    a = norm1.getAngle(norm2)
    if (a < 0.01) or (abs(a - math.pi) < 0.01):
        return True
    return False


def update_svg_cache(source, renderMode, showHidden, showFill, fillSpaces, joinArch, allOn, objs):
    """
    Returns None or cached SVG, clears shape cache if required
    """
    svgcache = None
    if hasattr(source, "Proxy"):
        if hasattr(source.Proxy, "svgcache") and source.Proxy.svgcache:
            # TODO check array bounds
            svgcache = source.Proxy.svgcache[0]
            # empty caches if we want to force-recalculate for certain properties
            if (
                source.Proxy.svgcache[1] != renderMode
                or source.Proxy.svgcache[2] != showHidden
                or source.Proxy.svgcache[3] != showFill
                or source.Proxy.svgcache[4] != fillSpaces
                or source.Proxy.svgcache[5] != joinArch
                or source.Proxy.svgcache[6] != allOn
                or source.Proxy.svgcache[7] != set(objs)
            ):
                svgcache = None
            if (
                source.Proxy.svgcache[4] != fillSpaces
                or source.Proxy.svgcache[5] != joinArch
                or source.Proxy.svgcache[6] != allOn
                or source.Proxy.svgcache[7] != set(objs)
            ):
                source.Proxy.shapecache = None
    return svgcache


def getSVG(
    source,
    renderMode="Wireframe",
    allOn=False,
    showHidden=False,
    scale=1,
    rotation=0,
    linewidth=1,
    lineColor=(0.0, 0.0, 0.0),
    fontsize=1,
    linespacing=None,
    showFill=False,
    fillColor=(1.0, 1.0, 1.0),
    techdraw=False,
    fillSpaces=False,
    cutlinewidth=0,
    joinArch=False,
):
    """
    Return an SVG fragment from an Arch SectionPlane or BuildingPart.

    allOn
        If it is `True`, all cut objects are shown, regardless of if they are
        visible or not.

    renderMode
        Can be `'Wireframe'` (default) or `'Solid'` to use the Arch solid
        renderer.

    showHidden
        If it is `True`, the hidden geometry above the section plane
        is shown in dashed line.

    showFill
        If it is `True`, the cut areas get filled with a pattern.

    lineColor
        Color of lines for the `renderMode` is `'Wireframe'`.

    fillColor
        If `showFill` is `True` and `renderMode` is `'Wireframe'`,
        the cut areas are filled with `fillColor`.

    fillSpaces
        If `True`, shows space objects as filled surfaces.
    """
    import Part

    objs, cutplane, onlySolids, clip, direction = getSectionData(source)
    if not objs:
        return ""
    if not allOn:
        objs = Draft.removeHidden(objs)

    # separate spaces and Draft objects
    spaces = []
    nonspaces = []
    drafts = []  # Only used for annotations.
    windows = []
    cutface = None
    for o in objs:
        if Draft.getType(o) == "Space":
            spaces.append(o)
        elif Draft.getType(o) in [
            "Dimension",
            "AngularDimension",
            "LinearDimension",
            "Annotation",
            "Label",
            "Text",
            "DraftText",
            "Axis",
        ]:
            if isOriented(o, cutplane):
                drafts.append(o)
        elif o.isDerivedFrom("App::DocumentObjectGroup"):
            # These will have been expanded by getSectionData already
            pass
        else:
            nonspaces.append(o)
        if Draft.getType(o.getLinkedObject()) == "Window":  # To support Link of Windows(Doors)
            windows.append(o)
    objs = nonspaces

    scaledLineWidth = linewidth / scale
    if renderMode in ["Coin", 2, "Coin mono", 3]:
        # don't scale linewidths in coin mode
        svgLineWidth = str(linewidth) + "px"
    else:
        svgLineWidth = str(scaledLineWidth) + "px"
    if cutlinewidth:
        scaledCutLineWidth = cutlinewidth / scale
        svgCutLineWidth = str(scaledCutLineWidth) + "px"
    else:
        st = params.get_param_arch("CutLineThickness")
        svgCutLineWidth = str(scaledLineWidth * st) + "px"
    yt = params.get_param_arch("SymbolLineThickness")
    svgSymbolLineWidth = str(linewidth * yt)
    hiddenPattern = params.get_param_arch("archHiddenPattern")
    svgHiddenPattern = hiddenPattern.replace(" ", "")
    # fillpattern = '<pattern id="sectionfill" patternUnits="userSpaceOnUse" patternTransform="matrix(5,0,0,5,0,0)"'
    # fillpattern += ' x="0" y="0" width="10" height="10">'
    # fillpattern += '<g>'
    # fillpattern += '<rect width="10" height="10" style="stroke:none; fill:#ffffff" /><path style="stroke:#000000; stroke-width:1" d="M0,0 l10,10" /></g></pattern>'
    svgLineColor = Draft.getrgb(lineColor)
    svg = ""
    # reading cached version
    svgcache = update_svg_cache(
        source, renderMode, showHidden, showFill, fillSpaces, joinArch, allOn, objs
    )
    should_update_svg_cache = False
    if showFill or not svgcache:
        should_update_svg_cache = True

    # generating SVG
    if renderMode in ["Coin", 2, "Coin mono", 3]:
        # render using a coin viewer
        if hasattr(source.ViewObject, "ViewData") and source.ViewObject.ViewData:
            cameradata = None  # getCameraData(source.ViewObject.ViewData)
        else:
            cameradata = None
        if should_update_svg_cache:
            if renderMode in ["Coin mono", 3]:
                svgcache = getCoinSVG(
                    cutplane, objs, cameradata, linewidth="SVGLINEWIDTH", facecolor="#ffffff"
                )
            else:
                svgcache = getCoinSVG(cutplane, objs, cameradata, linewidth="SVGLINEWIDTH")
    elif renderMode in ["Solid", 1]:
        if should_update_svg_cache:
            svgcache = ""
            # render using the Arch Vector Renderer
            import ArchVRM
            import WorkingPlane

            wp = WorkingPlane.PlaneBase()
            pl = FreeCAD.Placement(source.Placement)
            if source.ViewObject and hasattr(source.ViewObject, "CutMargin"):
                mv = pl.multVec(FreeCAD.Vector(0, 0, 1))
                mv.multiply(source.ViewObject.CutMargin)
                pl.move(mv)
            wp.align_to_rotation(pl.Rotation)
            # wp.inverse()
            render = ArchVRM.Renderer()
            render.setWorkingPlane(wp)
            render.addObjects(objs)
            if showHidden:
                render.cut(cutplane, showHidden)
            else:
                render.cut(cutplane)
            g = '<g transform="scale(1,-1)">\n'
            if hasattr(source.ViewObject, "RotateSolidRender"):
                if source.ViewObject.RotateSolidRender.Value != 0:
                    g = '<g transform="scale(1,-1) rotate('
                    g += str(source.ViewObject.RotateSolidRender.Value)
                    g += ')">\n'
            svgcache += g
            svgcache += render.getViewSVG(linewidth="SVGLINEWIDTH")
            # svgcache += fillpattern
            svgcache += render.getSectionSVG(linewidth="SVGCUTLINEWIDTH", fillpattern="#ffffff")
            if showHidden:
                svgcache += render.getHiddenSVG(linewidth="SVGLINEWIDTH")
            svgcache += "</g>\n"
            # print(render.info())
    else:
        # Wireframe (0) mode

        if (
            hasattr(source, "Proxy")
            and hasattr(source.Proxy, "shapecache")
            and source.Proxy.shapecache
        ):
            vshapes = source.Proxy.shapecache[0]
            hshapes = source.Proxy.shapecache[1]
            sshapes = source.Proxy.shapecache[2]
            cutface = source.Proxy.shapecache[3]
            # cutvolume = source.Proxy.shapecache[4] # Unused
            # invcutvolume = source.Proxy.shapecache[5] # Unused
            objectSshapes = source.Proxy.shapecache[6]
        else:
            if showFill:
                vshapes, hshapes, sshapes, cutface, cutvolume, invcutvolume, objectSshapes = (
                    getCutShapes(objs, cutplane, onlySolids, clip, joinArch, showHidden, True)
                )
            else:
                vshapes, hshapes, sshapes, cutface, cutvolume, invcutvolume = getCutShapes(
                    objs, cutplane, onlySolids, clip, joinArch, showHidden
                )
                objectSshapes = []
            source.Proxy.shapecache = [
                vshapes,
                hshapes,
                sshapes,
                cutface,
                cutvolume,
                invcutvolume,
                objectSshapes,
            ]

        if should_update_svg_cache:
            svgcache = ""
            # render using the TechDraw module
            import TechDraw
            import Part

            if vshapes:
                baseshape = Part.makeCompound(vshapes)
                style = {"stroke": "SVGLINECOLOR", "stroke-width": "SVGLINEWIDTH"}
                svgcache += TechDraw.projectToSVG(
                    baseshape,
                    direction,
                    hStyle=style,
                    h0Style=style,
                    h1Style=style,
                    vStyle=style,
                    v0Style=style,
                    v1Style=style,
                )
            if hshapes:
                hshapes = Part.makeCompound(hshapes)
                style = {
                    "stroke": "SVGLINECOLOR",
                    "stroke-width": "SVGLINEWIDTH",
                    "stroke-dasharray": "SVGHIDDENPATTERN",
                }
                svgcache += TechDraw.projectToSVG(
                    hshapes,
                    direction,
                    hStyle=style,
                    h0Style=style,
                    h1Style=style,
                    vStyle=style,
                    v0Style=style,
                    v1Style=style,
                )
            if sshapes:
                if showFill:
                    # svgcache += fillpattern
                    svgcache += '<g transform="rotate(180)">\n'
                    for o, shapes in objectSshapes:
                        for s in shapes:
                            if s.Edges:
                                objectFill = getFillForObject(o, fillColor, source)
                                # svg += Draft.get_svg(s,
                                #                      direction=direction.negative(),
                                #                      linewidth=0,
                                #                      fillstyle="sectionfill",
                                #                      color=(0,0,0))
                                # temporarily disabling fill patterns
                                svgcache += Draft.get_svg(
                                    s,
                                    linewidth=0,
                                    fillstyle=Draft.getrgb(objectFill, testbw=False),
                                    direction=direction.negative(),
                                    color=lineColor,
                                )
                    svgcache += "</g>\n"
                sshapes = Part.makeCompound(sshapes)
                style = {"stroke": "SVGLINECOLOR", "stroke-width": "SVGCUTLINEWIDTH"}
                svgcache += TechDraw.projectToSVG(
                    sshapes,
                    direction,
                    hStyle=style,
                    h0Style=style,
                    h1Style=style,
                    vStyle=style,
                    v0Style=style,
                    v1Style=style,
                )
    if should_update_svg_cache:
        if hasattr(source, "Proxy"):
            source.Proxy.svgcache = [
                svgcache,
                renderMode,
                showHidden,
                showFill,
                fillSpaces,
                joinArch,
                allOn,
                set(objs),
            ]

    svgcache = svgcache.replace("SVGLINECOLOR", svgLineColor)
    svgcache = svgcache.replace("SVGLINEWIDTH", svgLineWidth)
    svgcache = svgcache.replace("SVGHIDDENPATTERN", svgHiddenPattern)
    svgcache = svgcache.replace("SVGCUTLINEWIDTH", svgCutLineWidth)
    svg += svgcache

    if drafts:
        if not techdraw:
            svg += '<g transform="scale(1,-1)">'
        for d in drafts:
            svg += Draft.get_svg(
                d,
                scale=scale,
                linewidth=svgSymbolLineWidth,
                fontsize=fontsize,
                linespacing=linespacing,
                direction=direction,
                color=lineColor,
                techdraw=techdraw,
                rotation=rotation,
                override=False,
            )
        if not techdraw:
            svg += "</g>"

    if not cutface:
        # if we didn't calculate anything better, use the cutplane...
        cutface = cutplane

    # filter out spaces not cut by the source plane
    if cutface and spaces:
        spaces = [s for s in spaces if s.Shape.BoundBox.intersect(cutface.BoundBox)]
    if spaces:
        if not techdraw:
            svg += '<g transform="scale(1,-1)">'
        for s in spaces:
            svg += Draft.get_svg(
                s,
                scale=scale,
                linewidth=svgSymbolLineWidth,
                fontsize=fontsize,
                linespacing=linespacing,
                direction=direction,
                color=lineColor,
                techdraw=techdraw,
                rotation=rotation,
                fillspaces=fillSpaces,
            )
        if not techdraw:
            svg += "</g>"

    # add additional edge symbols from windows
    cutwindows = []
    if cutface and windows and BoundBoxValid(cutface.BoundBox):
        cutwindows = [
            w.Name
            for w in windows
            if BoundBoxValid(w.Shape.BoundBox) and w.Shape.BoundBox.intersect(cutface.BoundBox)
        ]
    if windows:
        sh = []
        for w in windows:
            if w.Name in cutwindows:
                wlo = w.getLinkedObject()  # To support Link of Windows(Doors)
                if hasattr(wlo, "SymbolPlan") and wlo.SymbolPlan:
                    if not hasattr(wlo.Proxy, "sshapes"):
                        wlo.Proxy.execute(wlo)
                    if hasattr(wlo.Proxy, "sshapes") and wlo.Proxy.sshapes:
                        c = Part.makeCompound(wlo.Proxy.sshapes)
                        c.Placement = w.Placement
                        sh.append(c)
        if sh:
            if not techdraw:
                svg += '<g transform="scale(1,-1)">'
            for s in sh:
                svg += Draft.get_svg(
                    s,
                    scale=scale,
                    linewidth=svgSymbolLineWidth,
                    fontsize=fontsize,
                    linespacing=linespacing,
                    fillstyle="none",
                    direction=direction,
                    color=lineColor,
                    techdraw=techdraw,
                    rotation=rotation,
                )
            if not techdraw:
                svg += "</g>"

    return svg


def BoundBoxValid(boundBox) -> bool:
    """Return true if boundBox has a non-zero volume"""
    return boundBox.XLength > 0 and boundBox.YLength > 0 and boundBox.ZLength > 0


def getDXF(obj):
    """Return a DXF representation from a TechDraw view."""
    allOn = getattr(obj, "AllOn", True)
    showHidden = getattr(obj, "ShowHidden", False)
    result = []
    import TechDraw
    import Part

    if not obj.Source:
        return result
    source = obj.Source
    objs, cutplane, onlySolids, clip, direction = getSectionData(source)
    if not objs:
        return result
    if not allOn:
        objs = Draft.removeHidden(objs)
    objs = [
        obj
        for obj in objs
        if (
            not obj.isDerivedFrom("Part::Part2DObject")
            and Draft.getType(obj)
            not in ["BezCurve", "BSpline", "Wire", "Annotation", "Dimension", "Space"]
        )
    ]
    vshapes, hshapes, sshapes, cutface, cutvolume, invcutvolume = getCutShapes(
        objs, cutplane, onlySolids, clip, False, showHidden
    )
    if vshapes:
        result.append(TechDraw.projectToDXF(Part.makeCompound(vshapes), direction))
    if sshapes:
        result.append(TechDraw.projectToDXF(Part.makeCompound(sshapes), direction))
    if hshapes:
        result.append(TechDraw.projectToDXF(Part.makeCompound(hshapes), direction))
    return result


def getCameraData(floatlist):
    """reconstructs a valid camera data string from stored values"""

    c = ""
    if len(floatlist) >= 12:
        d = floatlist
        camtype = "orthographic"
        if len(floatlist) == 13:
            if d[12] == 1:
                camtype = "perspective"
        if camtype == "orthographic":
            c = "#Inventor V2.1 ascii\n\n\nOrthographicCamera {\n  viewportMapping ADJUST_CAMERA\n  "
        else:
            c = "#Inventor V2.1 ascii\n\n\nPerspectiveCamera {\n  viewportMapping ADJUST_CAMERA\n  "
        c += "position " + str(d[0]) + " " + str(d[1]) + " " + str(d[2]) + "\n  "
        c += (
            "orientation "
            + str(d[3])
            + " "
            + str(d[4])
            + " "
            + str(d[5])
            + "  "
            + str(d[6])
            + "\n  "
        )
        c += "aspectRatio " + str(d[9]) + "\n  "
        c += "focalDistance " + str(d[10]) + "\n  "
        if camtype == "orthographic":
            c += "height " + str(d[11]) + "\n\n}\n"
        else:
            c += "heightAngle " + str(d[11]) + "\n\n}\n"
    return c


def getCoinSVG(cutplane, objs, cameradata=None, linewidth=0.2, singleface=False, facecolor=None):
    """Returns an SVG fragment generated from a coin view"""

    if not FreeCAD.GuiUp:
        return ""

    # do not allow concurrent runs
    # wait until the other rendering has finished
    global ISRENDERING
    while ISRENDERING:
        time.sleep(0.1)

    ISRENDERING = True

    # a name to save a temp file
    svgfile = tempfile.mkstemp(suffix=".svg")[1]

    # set object lighting to single face to get black fills
    # but this creates artifacts in svg output, triangulation gets visible...
    ldict = {}
    if singleface:
        for obj in objs:
            if hasattr(obj, "ViewObject") and hasattr(obj.ViewObject, "Lighting"):
                ldict[obj.Name] = obj.ViewObject.Lighting
                obj.ViewObject.Lighting = "One side"

    # get nodes to render
    root_node = coin.SoSeparator()
    boundbox = FreeCAD.BoundBox()
    for obj in objs:
        if hasattr(obj.ViewObject, "RootNode") and obj.ViewObject.RootNode:
            old_visibility = obj.ViewObject.isVisible()
            # ignore visibility as only visible objects are passed here
            obj.ViewObject.show()
            node_copy = obj.ViewObject.RootNode.copy()
            root_node.addChild(node_copy)
            if old_visibility:
                obj.ViewObject.show()
            else:
                obj.ViewObject.hide()

        if hasattr(obj, "Shape") and hasattr(obj.Shape, "BoundBox"):
            boundbox.add(obj.Shape.BoundBox)

    # reset lighting of objects
    if ldict:
        for obj in objs:
            if obj.Name in ldict:
                obj.ViewObject.Lighting = ldict[obj.Name]

    # create viewer
    view_window = FreeCADGui.createViewer()
    view_window_name = "Temp" + str(uuid.uuid4().hex[:8])
    view_window.setName(view_window_name)
    # disable animations to prevent a crash:
    # https://github.com/FreeCAD/FreeCAD/issues/24929
    view_window.setAnimationEnabled(False)
    inventor_view = view_window.getViewer()

    inventor_view.setBackgroundColor(1, 1, 1)
    view_window.redraw()

    # set clip plane
    clip = coin.SoClipPlane()
    norm = cutplane.normalAt(0, 0).negative()
    proj = DraftVecUtils.project(cutplane.CenterOfMass, norm)
    dist = proj.Length
    if proj.getAngle(norm) > 1:
        dist = -dist
    clip.on = True
    plane = coin.SbPlane(coin.SbVec3f(norm.x, norm.y, norm.z), dist)  # dir, position on dir
    clip.plane.setValue(plane)
    root_node.insertChild(clip, 0)

    # add white marker at scene bound box corner
    markervec = FreeCAD.Vector(10, 10, 10)
    a = cutplane.normalAt(0, 0).getAngle(markervec)
    if (a < 0.01) or (abs(a - math.pi) < 0.01):
        markervec = FreeCAD.Vector(10, -10, 10)
    boundbox.enlarge(10)  # so the marker don't overlap the objects
    sep = coin.SoSeparator()
    mat = coin.SoMaterial()
    mat.diffuseColor.setValue([1, 1, 1])
    sep.addChild(mat)
    coords = coin.SoCoordinate3()
    coords.point.setValues(
        [
            [boundbox.XMin, boundbox.YMin, boundbox.ZMin],
            [boundbox.XMin + markervec.x, boundbox.YMin + markervec.y, boundbox.ZMin + markervec.z],
        ]
    )
    sep.addChild(coords)
    lset = coin.SoIndexedLineSet()
    lset.coordIndex.setValues(0, 2, [0, 1])
    sep.addChild(lset)
    root_node.insertChild(sep, 0)

    # set scenegraph
    inventor_view.setSceneGraph(root_node)

    # set camera
    if cameradata:
        view_window.setCamera(cameradata)
    else:
        view_window.setCameraType("Orthographic")
        # rot = FreeCAD.Rotation(FreeCAD.Vector(0,0,1),cutplane.normalAt(0,0))
        vx = cutplane.Placement.Rotation.multVec(FreeCAD.Vector(1, 0, 0))
        vy = cutplane.Placement.Rotation.multVec(FreeCAD.Vector(0, 1, 0))
        vz = cutplane.Placement.Rotation.multVec(FreeCAD.Vector(0, 0, 1))
        rot = FreeCAD.Rotation(vx, vy, vz, "ZXY")
        view_window.setCameraOrientation(rot.Q)
    # this is needed to set correct focal depth, otherwise saving doesn't work properly
    view_window.fitAll()

    # save view
    # print("saving to",svgfile)
    view_window.saveVectorGraphic(svgfile, 1)  # number is pixel size

    # set linewidth placeholder
    f = open(svgfile, "r")
    svg = f.read()
    f.close()
    svg = svg.replace("stroke-width:1.0;", "stroke-width:" + str(linewidth) + ";")
    svg = svg.replace('stroke-width="1px', 'stroke-width="' + str(linewidth))

    # find marker and calculate scale factor and translation
    # <line x1="284.986" y1="356.166" x2="285.038" y2="356.166" stroke="#ffffff" stroke-width="1px" />
    factor = None
    trans = None
    import WorkingPlane

    wp = WorkingPlane.PlaneBase()
    wp.align_to_point_and_axis_svg(Vector(0, 0, 0), cutplane.normalAt(0, 0), 0)
    p = wp.get_local_coords(markervec)
    orlength = FreeCAD.Vector(p.x, p.y, 0).Length
    marker = re.findall(r"<line x1=.*?stroke=\"\#ffffff\".*?\/>", svg)
    if marker:
        marker = marker[0].split('"')
        x1 = float(marker[1])
        y1 = float(marker[3])
        x2 = float(marker[5])
        y2 = float(marker[7])
        p1 = FreeCAD.Vector(x1, y1, 0)
        p2 = FreeCAD.Vector(x2, y2, 0)
        factor = orlength / p2.sub(p1).Length
        if factor:
            orig = wp.get_local_coords(FreeCAD.Vector(boundbox.XMin, boundbox.YMin, boundbox.ZMin))
            orig = FreeCAD.Vector(orig.x, -orig.y, 0)
            scaledp1 = FreeCAD.Vector(p1.x * factor, p1.y * factor, 0)
            trans = orig.sub(scaledp1)
        # remove marker
        svg = re.sub(r"<line x1=.*?stroke=\"\#ffffff\".*?\/>", "", svg, count=1)

    # remove background rectangle
    svg = re.sub(r"<path.*?>", "", svg, count=1, flags=re.MULTILINE | re.DOTALL)

    # set face color to white
    if facecolor:
        res = re.findall(r"fill:(.*?); stroke:(.*?);", svg)
        pairs = []
        for pair in res:
            if (pair not in pairs) and (pair[0] == pair[1]) and (pair[0] not in ["#0a0a0a"]):
                # coin seems to be rendering a lot of lines as thin triangles with the #0a0a0a color...
                pairs.append(pair)
        for pair in pairs:
            svg = re.sub(
                r"fill:" + pair[0] + "; stroke:" + pair[1] + ";",
                "fill:" + facecolor + "; stroke:" + facecolor + ";",
                svg,
            )

    # embed everything in a scale group and scale the viewport
    if factor:
        if trans:
            svg = svg.replace(
                "<g>",
                '<g transform="translate('
                + str(trans.x)
                + " "
                + str(trans.y)
                + ") scale("
                + str(factor)
                + ","
                + str(factor)
                + ')">\n<g>',
                1,
            )
        else:
            svg = svg.replace(
                "<g>", '<g transform="scale(' + str(factor) + "," + str(factor) + ')">\n<g>', 1
            )
        svg = svg.replace("</svg>", "</g>\n</svg>")

    # trigger viewer close
    QtCore.QTimer.singleShot(1, lambda: closeViewer(view_window_name))

    # strip svg tags (needed for TD Arch view)
    svg = re.sub(r"<\?xml.*?>", "", svg, flags=re.MULTILINE | re.DOTALL)
    svg = re.sub(r"<svg.*?>", "", svg, flags=re.MULTILINE | re.DOTALL)
    svg = re.sub(r"<\/svg>", "", svg, flags=re.MULTILINE | re.DOTALL)

    ISRENDERING = False

    return svg


def closeViewer(name):
    """Close temporary viewers"""

    mw = FreeCADGui.getMainWindow()
    for sw in mw.findChildren(QtGui.QMdiSubWindow):
        if sw.windowTitle() == name:
            sw.close()


class _SectionPlane:
    "A section plane object"

    def __init__(self, obj):
        obj.Proxy = self
        self.Type = "SectionPlane"
        self.setProperties(obj)

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Placement" in pl:
            obj.addProperty(
                "App::PropertyPlacement",
                "Placement",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The placement of this object"),
                locked=True,
            )
        if not "Shape" in pl:
            obj.addProperty(
                "Part::PropertyPartShape",
                "Shape",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The shape of this object"),
                locked=True,
            )
        if not "Objects" in pl:
            obj.addProperty(
                "App::PropertyLinkList",
                "Objects",
                "SectionPlane",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The objects that must be considered by this section plane. Empty means the whole document.",
                ),
                locked=True,
            )
        if not "OnlySolids" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "OnlySolids",
                "SectionPlane",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If false, non-solids will be cut too, with possible wrong results.",
                ),
                locked=True,
            )
            obj.OnlySolids = True
        if not "Clip" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "Clip",
                "SectionPlane",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If True, resulting views will be clipped to the section plane area.",
                ),
                locked=True,
            )
        if not "UseMaterialColorForFill" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "UseMaterialColorForFill",
                "SectionPlane",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If true, the color of the objects material will be used to fill cut areas.",
                ),
                locked=True,
            )
            obj.UseMaterialColorForFill = False
        if not "Depth" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Depth",
                "SectionPlane",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Geometry further than this value will be cut off. Keep zero for unlimited.",
                ),
                locked=True,
            )

    def onDocumentRestored(self, obj):

        self.setProperties(obj)

    def execute(self, obj):
        import math
        import Part

        l = 1
        h = 1
        if obj.ViewObject:
            if hasattr(obj.ViewObject, "DisplayLength"):
                l = obj.ViewObject.DisplayLength.Value
                h = obj.ViewObject.DisplayHeight.Value
            elif hasattr(obj.ViewObject, "DisplaySize"):
                # old objects
                l = obj.ViewObject.DisplaySize.Value
                h = obj.ViewObject.DisplaySize.Value
        if not l:
            l = 1
        if not h:
            h = 1
        p = Part.makePlane(l, h, Vector(l / 2, -h / 2, 0), Vector(0, 0, -1))
        # make sure the normal direction is pointing outwards, you never know what OCC will decide...
        # Apply the object's placement to the new plane first.
        p.Placement = obj.Placement

        # Now, check if the resulting plane's normal matches the placement's intended direction.
        # This robustly handles all rotation angles and potential OCC inconsistencies.
        target_normal = obj.Placement.Rotation.multVec(FreeCAD.Vector(0, 0, 1))
        if p.normalAt(0, 0).getAngle(target_normal) > math.pi / 2:
            p.reverse()
        obj.Shape = p
        self.svgcache = None
        self.shapecache = None

    def getNormal(self, obj):

        return obj.Shape.Faces[0].normalAt(0, 0)

    def dumps(self):

        return None

    def loads(self, state):

        self.Type = "SectionPlane"


class _ViewProviderSectionPlane:
    "A View Provider for Section Planes"

    def __init__(self, vobj):

        vobj.Proxy = self
        self.setProperties(vobj)

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        d = 0
        if "DisplaySize" in pl:
            d = vobj.DisplaySize.Value
            vobj.removeProperty("DisplaySize")
        if not "DisplayLength" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "DisplayLength",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The display length of this section plane"),
                locked=True,
            )
            if d:
                vobj.DisplayLength = d
            else:
                vobj.DisplayLength = 1000
        if not "DisplayHeight" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "DisplayHeight",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The display height of this section plane"),
                locked=True,
            )
            if d:
                vobj.DisplayHeight = d
            else:
                vobj.DisplayHeight = 1000
        if not "ArrowSize" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "ArrowSize",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The size of the arrows of this section plane"),
                locked=True,
            )
            vobj.ArrowSize = 50
        if not "Transparency" in pl:
            vobj.addProperty(
                "App::PropertyPercent",
                "Transparency",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The transparency of this object"),
                locked=True,
            )
            vobj.Transparency = 85
        if not "LineWidth" in pl:
            vobj.addProperty(
                "App::PropertyFloat",
                "LineWidth",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The line width of this object"),
                locked=True,
            )
            vobj.LineWidth = 1
        if not "CutDistance" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "CutDistance",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "Show the cut in the 3D view"),
                locked=True,
            )
        if not "LineColor" in pl:
            vobj.addProperty(
                "App::PropertyColor",
                "LineColor",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The color of this object"),
                locked=True,
            )
            vobj.LineColor = ArchCommands.getDefaultColor("Helpers")
        if not "CutView" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "CutView",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "Show the cut in the 3D view"),
                locked=True,
            )
        if not "CutMargin" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "CutMargin",
                "SectionPlane",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The distance between the cut plane and the actual view cut (keep this a very small value but not zero)",
                ),
                locked=True,
            )
            vobj.CutMargin = 1
        if not "ShowLabel" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowLabel",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "Show the label in the 3D view"),
                locked=True,
            )
        if not "FontName" in pl:
            vobj.addProperty(
                "App::PropertyFont",
                "FontName",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The name of the font"),
                locked=True,
            )
            vobj.FontName = params.get_param("textfont")
        if not "FontSize" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "FontSize",
                "SectionPlane",
                QT_TRANSLATE_NOOP("App::Property", "The size of the text font"),
                locked=True,
            )
            vobj.FontSize = params.get_param("textheight") * params.get_param(
                "DefaultAnnoScaleMultiplier"
            )

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_SectionPlane_Tree.svg"

    def claimChildren(self):
        # buggy at the moment so it's disabled - it will for ex. swallow a building object directly at the root of the document...
        # if hasattr(self,"Object") and hasattr(self.Object,"Objects"):
        #    return self.Object.Objects
        return []

    def attach(self, vobj):

        self.Object = vobj.Object
        self.clip = None
        self.main_transform = gui_utils.find_coin_node(vobj.RootNode, coin.SoTransform)
        self.mat1 = coin.SoMaterial()
        self.mat2 = coin.SoMaterial()
        self.fcoords = coin.SoCoordinate3()
        # fs = coin.SoType.fromName("SoBrepFaceSet").createInstance() # this causes a FreeCAD freeze for me
        fs = coin.SoIndexedFaceSet()
        fs.coordIndex.setValues(0, 7, [0, 1, 2, -1, 0, 2, 3])
        self.drawstyle = coin.SoDrawStyle()
        self.drawstyle.style = coin.SoDrawStyle.LINES
        self.lcoords = coin.SoCoordinate3()
        import PartGui  # Required for "SoBrepEdgeSet" (because a SectionPlane is not a Part::FeaturePython object).

        ls = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        ls.coordIndex.setValues(
            0,
            57,
            [
                0,
                1,
                -1,
                2,
                3,
                4,
                5,
                -1,
                6,
                7,
                8,
                9,
                -1,
                10,
                11,
                -1,
                12,
                13,
                14,
                15,
                -1,
                16,
                17,
                18,
                19,
                -1,
                20,
                21,
                -1,
                22,
                23,
                24,
                25,
                -1,
                26,
                27,
                28,
                29,
                -1,
                30,
                31,
                -1,
                32,
                33,
                34,
                35,
                -1,
                36,
                37,
                38,
                39,
                -1,
                40,
                41,
                42,
                43,
                44,
            ],
        )
        self.txtcoords = coin.SoTransform()
        self.txtfont = coin.SoFont()
        self.txtfont.name = ""
        self.txt = coin.SoAsciiText()
        self.txt.justification = coin.SoText2.LEFT
        self.txt.string.setValue(" ")
        sep = coin.SoSeparator()
        psep = coin.SoSeparator()
        fsep = coin.SoSeparator()
        tsep = coin.SoSeparator()
        fsep.addChild(self.mat2)
        fsep.addChild(self.fcoords)
        fsep.addChild(fs)
        psep.addChild(self.mat1)
        psep.addChild(self.drawstyle)
        psep.addChild(self.lcoords)
        psep.addChild(ls)
        tsep.addChild(self.mat1)
        tsep.addChild(self.txtcoords)
        tsep.addChild(self.txtfont)
        tsep.addChild(self.txt)
        sep.addChild(fsep)
        sep.addChild(psep)
        sep.addChild(tsep)
        vobj.addDisplayMode(sep, "Default")
        self.onChanged(vobj, "DisplayLength")
        self.onChanged(vobj, "LineColor")
        self.onChanged(vobj, "Transparency")
        self.onChanged(vobj, "CutView")

    def getDisplayModes(self, vobj):

        return ["Default"]

    def getDefaultDisplayMode(self):

        return "Default"

    def setDisplayMode(self, mode):

        return mode

    def updateData(self, obj, prop):
        vobj = obj.ViewObject
        if prop in ["Placement"]:
            self.onChanged(vobj, "DisplayLength")

            # Defer the clipping plane update until after the current event
            # loop finishes. This ensures the scene graph has been updated with the
            # new placement before we try to recalculate the clip plane.
            if vobj and hasattr(vobj, "CutView") and vobj.CutView:
                from PySide import QtCore

                # We use a lambda to pass the vobj argument to the delayed function.
                QtCore.QTimer.singleShot(0, lambda: self.refreshCutView(vobj))
        elif prop == "Label":
            if hasattr(obj.ViewObject, "ShowLabel") and obj.ViewObject.ShowLabel:
                self.txt.string = obj.Label
        return

    def refreshCutView(self, vobj):
        """
        Forces a refresh of the SoClipPlane by toggling the CutView property.
        This is called with a delay to ensure the object's placement is up-to-date.
        """
        if vobj and hasattr(vobj, "CutView") and vobj.CutView:
            vobj.CutView = False
            vobj.CutView = True

    def onChanged(self, vobj, prop):

        if prop == "LineColor":
            if hasattr(vobj, "LineColor"):
                l = vobj.LineColor
                self.mat1.diffuseColor.setValue([l[0], l[1], l[2]])
                self.mat2.diffuseColor.setValue([l[0], l[1], l[2]])
        elif prop == "Transparency":
            if hasattr(vobj, "Transparency"):
                self.mat2.transparency.setValue(vobj.Transparency / 100.0)
        elif prop in ["DisplayLength", "DisplayHeight", "ArrowSize"]:
            # for IFC objects: propagate to the object
            if prop in ["DisplayLength", "DisplayHeight"]:
                if hasattr(vobj.Object.Proxy, "onChanged"):
                    vobj.Object.Proxy.onChanged(vobj.Object, prop)
            if hasattr(vobj, "DisplayLength") and hasattr(vobj, "DisplayHeight"):
                ld = vobj.DisplayLength.Value / 2
                hd = vobj.DisplayHeight.Value / 2
            elif hasattr(vobj, "DisplaySize"):
                # old objects
                ld = vobj.DisplaySize.Value / 2
                hd = vobj.DisplaySize.Value / 2
            else:
                ld = 1
                hd = 1
            if hasattr(vobj, "ArrowSize"):
                l1 = vobj.ArrowSize.Value if vobj.ArrowSize.Value > 0 else 0.1
            else:
                l1 = 0.1
            l2 = l1 / 3
            pl = vobj.Object.Placement
            self.main_transform.translation.setValue(pl.Base)
            self.main_transform.rotation = coin.SbRotation(pl.Rotation.Q)
            verts = []
            fverts = []
            for v in [[-ld, -hd], [ld, -hd], [ld, hd], [-ld, hd]]:
                p1 = Vector(v[0], v[1], 0)
                p2 = Vector(v[0], v[1], -l1)
                p3 = Vector(v[0] - l2, v[1], -l1 + l2)
                p4 = Vector(v[0] + l2, v[1], -l1 + l2)
                p5 = Vector(v[0], v[1] - l2, -l1 + l2)
                p6 = Vector(v[0], v[1] + l2, -l1 + l2)
                fverts.append(p1)
                verts.extend([p1, p2])
                verts.extend([p2, p3, p4, p2])
                verts.extend([p2, p5, p6, p2])
            verts.extend(fverts + [fverts[0]])
            p7 = Vector(-ld + l2, -hd + l2, 0)  # text pos
            self.lcoords.point.setValues(verts)
            self.fcoords.point.setValues(fverts)
            self.txtcoords.translation.setValue(p7)
            # self.txtfont.size = l1
        elif prop == "LineWidth":
            self.drawstyle.lineWidth = vobj.LineWidth
        elif prop in ["CutView", "CutMargin"]:
            if (
                hasattr(vobj, "CutView")
                and FreeCADGui.ActiveDocument.ActiveView
                and hasattr(FreeCADGui.ActiveDocument.ActiveView, "getSceneGraph")
            ):
                sg = FreeCADGui.ActiveDocument.ActiveView.getSceneGraph()
                if vobj.CutView:
                    if self.clip:
                        sg.removeChild(self.clip)
                        self.clip = None
                    for o in Draft.get_group_contents(vobj.Object.Objects, walls=True):
                        if hasattr(o.ViewObject, "Lighting"):
                            o.ViewObject.Lighting = "One side"
                    self.clip = coin.SoClipPlane()
                    self.clip.on.setValue(True)
                    norm = vobj.Object.Proxy.getNormal(vobj.Object)
                    mp = vobj.Object.Shape.CenterOfMass
                    mp = DraftVecUtils.project(mp, norm)
                    dist = mp.Length  # - 0.1 # to not clip exactly on the section object
                    norm = norm.negative()
                    marg = 1
                    if hasattr(vobj, "CutMargin"):
                        marg = vobj.CutMargin.Value
                    if mp.getAngle(norm) > 1:
                        dist += marg
                        dist = -dist
                    else:
                        dist -= marg
                    plane = coin.SbPlane(coin.SbVec3f(norm.x, norm.y, norm.z), dist)
                    self.clip.plane.setValue(plane)
                    sg.insertChild(self.clip, 0)
                else:
                    if self.clip:
                        sg.removeChild(self.clip)
                        self.clip = None
        elif prop == "ShowLabel":
            if vobj.ShowLabel:
                self.txt.string = vobj.Object.Label or " "
            else:
                self.txt.string = " "
        elif prop == "FontName":
            if hasattr(self, "txtfont") and hasattr(vobj, "FontName"):
                if vobj.FontName:
                    self.txtfont.name = vobj.FontName
                else:
                    self.txtfont.name = ""
        elif prop == "FontSize":
            if hasattr(self, "txtfont") and hasattr(vobj, "FontSize"):
                self.txtfont.size = vobj.FontSize.Value
        return

    def dumps(self):

        return None

    def loads(self, state):

        return None

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        taskd = SectionPlaneTaskPanel()
        taskd.obj = vobj.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        FreeCADGui.Control.closeDialog()
        return True

    def doubleClicked(self, vobj):
        self.edit()
        return True

    def setupContextMenu(self, vobj, menu):
        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        actionEdit = QtGui.QAction(translate("Arch", "Edit"), menu)
        QtCore.QObject.connect(actionEdit, QtCore.SIGNAL("triggered()"), self.edit)
        menu.addAction(actionEdit)

        actionToggleCutview = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_Edit.svg"), translate("Arch", "Toggle Cutview"), menu
        )
        actionToggleCutview.triggered.connect(lambda: self.toggleCutview(vobj))
        menu.addAction(actionToggleCutview)

    def edit(self):
        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)

    def toggleCutview(self, vobj):
        vobj.CutView = not vobj.CutView


class SectionPlaneTaskPanel:
    """A TaskPanel for all the section plane object"""

    def __init__(self):

        # the panel has a tree widget that contains categories
        # for the subcomponents, such as additions, subtractions.
        # the categories are shown only if they are not empty.

        self.obj = None

        # Create the first box for object scope
        self.scope_widget = QtGui.QWidget()
        scope_layout = QtGui.QGridLayout(self.scope_widget)

        self.title = QtGui.QLabel(self.scope_widget)
        scope_layout.addWidget(self.title, 0, 0, 1, 2)

        # tree
        self.tree = QtGui.QTreeWidget(self.scope_widget)
        scope_layout.addWidget(self.tree, 1, 0, 1, 2)
        self.tree.setColumnCount(1)
        self.tree.header().hide()

        # add / remove buttons
        self.addButton = QtGui.QPushButton(self.scope_widget)
        self.addButton.setIcon(QtGui.QIcon(":/icons/Arch_Add.svg"))
        scope_layout.addWidget(self.addButton, 2, 0, 1, 1)

        self.delButton = QtGui.QPushButton(self.scope_widget)
        self.delButton.setIcon(QtGui.QIcon(":/icons/Arch_Remove.svg"))
        scope_layout.addWidget(self.delButton, 2, 1, 1, 1)
        self.delButton.setEnabled(False)

        # Create the second box for tools
        self.tools_widget = QtGui.QWidget()
        tools_layout = QtGui.QVBoxLayout(self.tools_widget)

        # Cut View toggle button
        self.cutViewButton = QtGui.QPushButton(self.tools_widget)
        self.cutViewButton.setIcon(QtGui.QIcon(":/icons/Arch_CutPlane.svg"))
        self.cutViewButton.setObjectName("cutViewButton")
        self.cutViewButton.setCheckable(True)
        QtCore.QObject.connect(
            self.cutViewButton, QtCore.SIGNAL("toggled(bool)"), self.toggleCutView
        )
        tools_layout.addWidget(self.cutViewButton)

        # rotate / resize buttons
        self.rotation_label = QtGui.QLabel(self.tools_widget)
        tools_layout.addWidget(self.rotation_label)

        rotation_layout = QtGui.QHBoxLayout()
        self.rotateXButton = QtGui.QPushButton(self.tools_widget)
        self.rotateYButton = QtGui.QPushButton(self.tools_widget)
        self.rotateZButton = QtGui.QPushButton(self.tools_widget)
        rotation_layout.addWidget(self.rotateXButton)
        rotation_layout.addWidget(self.rotateYButton)
        rotation_layout.addWidget(self.rotateZButton)
        tools_layout.addLayout(rotation_layout)

        size_pos_layout = QtGui.QHBoxLayout()
        self.resizeButton = QtGui.QPushButton(self.tools_widget)
        self.recenterButton = QtGui.QPushButton(self.tools_widget)
        size_pos_layout.addWidget(self.resizeButton)
        size_pos_layout.addWidget(self.recenterButton)
        tools_layout.addLayout(size_pos_layout)

        QtCore.QObject.connect(self.addButton, QtCore.SIGNAL("clicked()"), self.addElement)
        QtCore.QObject.connect(self.delButton, QtCore.SIGNAL("clicked()"), self.removeElement)
        QtCore.QObject.connect(self.rotateXButton, QtCore.SIGNAL("clicked()"), self.rotateX)
        QtCore.QObject.connect(self.rotateYButton, QtCore.SIGNAL("clicked()"), self.rotateY)
        QtCore.QObject.connect(self.rotateZButton, QtCore.SIGNAL("clicked()"), self.rotateZ)
        QtCore.QObject.connect(self.resizeButton, QtCore.SIGNAL("clicked()"), self.resize)
        QtCore.QObject.connect(self.recenterButton, QtCore.SIGNAL("clicked()"), self.recenter)
        QtCore.QObject.connect(self.tree, QtCore.SIGNAL("itemSelectionChanged()"), self.onTreeClick)

        self.form = [self.scope_widget, self.tools_widget]
        self.update()

    def isAllowedAlterSelection(self):
        return True

    def isAllowedAlterView(self):
        return True

    def getStandardButtons(self):
        return QtGui.QDialogButtonBox.Close

    def getIcon(self, obj):
        if hasattr(obj.ViewObject, "Proxy"):
            return QtGui.QIcon(obj.ViewObject.Proxy.getIcon())
        elif obj.isDerivedFrom("Sketcher::SketchObject"):
            return QtGui.QIcon(":/icons/Sketcher_Sketch.svg")
        elif obj.isDerivedFrom("App::DocumentObjectGroup"):
            return QtGui.QApplication.style().standardIcon(QtGui.QStyle.SP_DirIcon)
        elif hasattr(obj.ViewObject, "Icon"):
            return QtGui.QIcon(obj.ViewObject.Icon)
        return QtGui.QIcon(":/icons/Part_3D_object.svg")

    def update(self):
        "fills the treewidget"
        self.tree.clear()
        if self.obj:
            for o in self.obj.Objects:
                item = QtGui.QTreeWidgetItem(self.tree)
                item.setText(0, o.Label)
                item.setToolTip(0, o.Name)
                item.setIcon(0, self.getIcon(o))
            if self.obj.ViewObject and hasattr(self.obj.ViewObject, "CutView"):
                self.cutViewButton.setChecked(self.obj.ViewObject.CutView)
        self.retranslateUi()

    def addElement(self):
        if self.obj:
            added = False
            for o in FreeCADGui.Selection.getSelection():
                if o != self.obj:
                    ArchComponent.addToComponent(self.obj, o, "Objects")
                    added = True
            if added:
                self.update()
            else:
                FreeCAD.Console.PrintWarning(
                    "Select objects in the 3D view or in the model tree before pressing the button\n"
                )

    def removeElement(self):
        if self.obj:
            it = self.tree.currentItem()
            if it:
                comp = FreeCAD.ActiveDocument.getObject(str(it.toolTip(0)))
                ArchComponent.removeFromComponent(self.obj, comp)
            self.update()

    def rotate(self, axis):
        if self.obj and self.obj.Shape and self.obj.Shape.Faces:
            face = self.obj.Shape.copy()
            import Part

            local_axis = self.obj.Placement.Rotation.multVec(axis)
            face.rotate(self.obj.Placement.Base, local_axis, 90)
            self.obj.Placement = face.Placement
            self.obj.Proxy.execute(self.obj)

    def rotateX(self):
        self.rotate(FreeCAD.Vector(1, 0, 0))

    def rotateY(self):
        self.rotate(FreeCAD.Vector(0, 1, 0))

    def rotateZ(self):
        self.rotate(FreeCAD.Vector(0, 0, 1))

    def getBB(self):
        bb = FreeCAD.BoundBox()
        if self.obj:
            for o in Draft.get_group_contents(self.obj.Objects):
                if hasattr(o, "Shape") and hasattr(o.Shape, "BoundBox"):
                    bb.add(o.Shape.BoundBox)
        return bb

    def resize(self):
        if self.obj and self.obj.ViewObject:
            bb = self.getBB()
            n = self.obj.Proxy.getNormal(self.obj)
            margin = bb.XLength * 0.1
            if (n.getAngle(FreeCAD.Vector(1, 0, 0)) < 0.1) or (
                n.getAngle(FreeCAD.Vector(-1, 0, 0)) < 0.1
            ):
                self.obj.ViewObject.DisplayLength = bb.YLength + margin
                self.obj.ViewObject.DisplayHeight = bb.ZLength + margin
            elif (n.getAngle(FreeCAD.Vector(0, 1, 0)) < 0.1) or (
                n.getAngle(FreeCAD.Vector(0, -1, 0)) < 0.1
            ):
                self.obj.ViewObject.DisplayLength = bb.XLength + margin
                self.obj.ViewObject.DisplayHeight = bb.ZLength + margin
            elif (n.getAngle(FreeCAD.Vector(0, 0, 1)) < 0.1) or (
                n.getAngle(FreeCAD.Vector(0, 0, -1)) < 0.1
            ):
                self.obj.ViewObject.DisplayLength = bb.XLength + margin
                self.obj.ViewObject.DisplayHeight = bb.YLength + margin
            self.obj.Proxy.execute(self.obj)

    def recenter(self):
        if self.obj:
            self.obj.Placement.Base = self.getBB().Center

    def onTreeClick(self):
        if self.tree.selectedItems():
            self.delButton.setEnabled(True)
        else:
            self.delButton.setEnabled(False)

    def accept(self):
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def toggleCutView(self, checked):
        if self.obj and self.obj.ViewObject and hasattr(self.obj.ViewObject, "CutView"):
            self.obj.ViewObject.CutView = checked

    def retranslateUi(self):
        self.scope_widget.setWindowTitle(QtGui.QApplication.translate("Arch", "Scope", None))
        self.tools_widget.setWindowTitle(
            QtGui.QApplication.translate("Arch", "Placement and Visuals", None)
        )
        self.title.setText(
            QtGui.QApplication.translate("Arch", "Objects seen by this section plane", None)
        )
        self.delButton.setText(QtGui.QApplication.translate("Arch", "Remove", None))
        self.delButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Removes highlighted objects from the list above", None
            )
        )
        self.addButton.setText(QtGui.QApplication.translate("Arch", "Add Selected", None))
        self.addButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Adds selected objects to the scope of this section plane", None
            )
        )
        self.cutViewButton.setText(QtGui.QApplication.translate("Arch", "Cut View", None))
        self.cutViewButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch",
                "Creates a live cut in the 3D view, hiding geometry on one side of the plane to see inside your model",
                None,
            )
        )
        self.rotation_label.setText(QtGui.QApplication.translate("Arch", "Rotate by 90°", None))
        self.rotateXButton.setText(QtGui.QApplication.translate("Arch", "Rotate X", None))
        self.rotateXButton.setToolTip(
            QtGui.QApplication.translate("Arch", "Rotates the plane around its local X-axis", None)
        )
        self.rotateYButton.setText(QtGui.QApplication.translate("Arch", "Rotate Y", None))
        self.rotateYButton.setToolTip(
            QtGui.QApplication.translate("Arch", "Rotates the plane around its local Y-axis", None)
        )
        self.rotateZButton.setText(QtGui.QApplication.translate("Arch", "Rotate Z", None))
        self.rotateZButton.setToolTip(
            QtGui.QApplication.translate("Arch", "Rotates the plane around its local Z-axis", None)
        )
        self.resizeButton.setText(QtGui.QApplication.translate("Arch", "Resize to Fit", None))
        self.resizeButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Resizes the plane to fit the objects in the list above", None
            )
        )
        self.recenterButton.setText(QtGui.QApplication.translate("Arch", "Recenter Plane", None))
        self.recenterButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Centers the plane on the objects in the list above", None
            )
        )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Axis-related commands"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Axis:
    "the Arch Axis command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Axis",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Axis", "Axis"),
            "Accel": "A, X",
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Axis", "Creates a set of axes"),
        }

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Axis"))
        FreeCADGui.addModule("Arch")

        FreeCADGui.doCommand("Arch.makeAxis()")
        FreeCAD.ActiveDocument.commitTransaction()

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


class Arch_AxisSystem:
    "the Arch Axis System command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Axis_System",
            "MenuText": QT_TRANSLATE_NOOP("Arch_AxisSystem", "Axis System"),
            "Accel": "X, S",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_AxisSystem", "Creates an axis system from a set of axes"
            ),
        }

    def Activated(self):

        import Draft

        if FreeCADGui.Selection.getSelection():
            s = "["
            for o in FreeCADGui.Selection.getSelection():
                if Draft.getType(o) != "Axis":
                    FreeCAD.Console.PrintError(
                        translate("Arch", "Only axes must be selected") + "\n"
                    )
                    return
                s += "FreeCAD.ActiveDocument." + o.Name + ","
            s += "]"
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Axis System"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand("Arch.makeAxisSystem(" + s + ")")
            FreeCAD.ActiveDocument.commitTransaction()
        else:
            FreeCAD.Console.PrintError(translate("Arch", "Select at least one axis") + "\n")

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


class Arch_Grid:
    "the Arch Grid command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Grid",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Grid", "Grid"),
            "Accel": "A, X",
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Grid", "Creates a customizable grid object"),
        }

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Grid"))
        FreeCADGui.addModule("Arch")

        FreeCADGui.doCommand("Arch.makeGrid()")
        FreeCAD.ActiveDocument.commitTransaction()

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


class Arch_AxisTools:
    """The Axis tools group command"""

    def GetCommands(self):
        return tuple(["Arch_Axis", "Arch_AxisSystem", "Arch_Grid"])

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("Arch_AxisTools", "Axis Tools"),
            "ToolTip": QT_TRANSLATE_NOOP("Arch_AxisTools", "Axis tools"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


FreeCADGui.addCommand("Arch_Axis", Arch_Axis())
FreeCADGui.addCommand("Arch_AxisSystem", Arch_AxisSystem())
FreeCADGui.addCommand("Arch_Grid", Arch_Grid())
FreeCADGui.addCommand("Arch_AxisTools", Arch_AxisTools())
